"""
Create a NeXus file for time-resolved collections on I19-2 using parameters passed from GDA.
"""
from __future__ import annotations

import glob
import logging
from collections import namedtuple
from pathlib import Path
from typing import Tuple

from hdf5plugin import Bitshuffle  # noqa: F401

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write import calculate_scan_range
from ..nxs_write.NexusWriter import call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry
from ..tools.VDS_tools import image_vds_writer
from . import eiger_meta_links as dset_links
from .GDAtools.ExtendedRequest import ExtendedRequestIO
from .GDAtools.GDAjson2params import (
    read_detector_params_from_json,
    read_geometry_from_json,
)
from .I19_2_params import (
    eiger4M_module,
    eiger4M_params,
    goniometer_axes,
    source,
    tristan10M_module,
    tristan10M_params,
)

import h5py  # isort: skip


# Define a logger object and a formatter
logger = logging.getLogger("nexgen.I19-2_NeXus_gda")

# Tristan mask and flatfield files
maskfile = "Tristan10M_mask_with_spec.h5"
flatfieldfile = "Tristan10M_flat_field_coeff_with_Mo_17.479keV.h5"

tr_collect = namedtuple(
    "tr_collect",
    [
        "meta_file",
        "xml_file",
        "detector_name",
        "exposure_time",
        "wavelength",
        "beam_center",
        "start_time",
        "stop_time",
        "geometry_json",  # Define these 2 as None
        "detector_json",
    ],
)

tr_collect.__doc__ = (
    """Information extracted from GDA containing collection parameters."""
)
tr_collect.meta_file.__doc__ = "Path to _meta.h5 file."
tr_collect.xml_file.__doc__ = "Path to GDA-generated xml file."
tr_collect.detector_name.__doc__ = "Name of the detector in use for current experiment."
tr_collect.exposure_time.__doc__ = "Exposure time, in s."
tr_collect.wavelength.__doc__ = "Incident beam wavelength, in A."
tr_collect.beam_center.__doc__ = "Beam center (x,y) position, in pixels."
tr_collect.start_time.__doc__ = "Collection start time."
tr_collect.stop_time.__doc__ = "Collection end time."
tr_collect.geometry_json.__doc__ = (
    "Path to GDA-generated JSON file describing the beamline geometry."
)
tr_collect.detector_json.__doc__ = (
    "Path to GDA-generated JSON file describing the detector."
)

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize dictionaries
goniometer = {}  # goniometer_axes
detector = {}  # tristan10M_params
module = {}
beam = {"flux": None}
attenuator = {}


def read_from_xml(xmlfile: Path | str, detector_name: str):
    """
    Extract information about the collection and the beamline contained in the xml file.

    Args:
        xmlfile (Path | str): Path to xml file.
        detector_name (str): Detector in use for the current collection

    Returns:
        scan_axis (str): Name of the rotation scan axis
        pos (Dict): Dictionary containing the (start,end,increment) values for each goniometer axis.
        num (int): Number of images written.
    """
    ecr = ExtendedRequestIO(xmlfile)

    # Attenuator
    attenuator["transmission"] = ecr.getTransmission()

    # Detector [2theta, det_z]
    if "tristan" in detector_name.lower():
        detector["starts"] = [0.0, ecr.getSampleDetectorDistance()]
    else:
        detector["starts"] = [ecr.getTwoTheta(), ecr.getSampleDetectorDistance()]

    # Goniometer
    osc_seq = ecr.getOscillationSequence()
    # Find scan range
    if osc_seq["range"] == 0.0:
        scan_range = (osc_seq["start"], osc_seq["start"])
        num = osc_seq["number_of_images"]
    else:
        start = osc_seq["start"]
        num = osc_seq["number_of_images"]
        stop = start + num * osc_seq["range"]
        scan_range = (start, stop)
    # Determine scan axis
    if ecr.getAxisChoice() == "omega":
        scan_axis = "omega"
        omega_pos = (*scan_range, osc_seq["range"])  # 0.0)
        phi_pos = (*2 * (ecr.getOtherAxis(),), 0.0)
    else:
        scan_axis = "phi"
        phi_pos = (*scan_range, osc_seq["range"])  # 0.0)
        omega_pos = (*2 * (ecr.getOtherAxis(),), 0.0)
    kappa_pos = (*2 * (ecr.getKappa(),), 0.0)

    pos = {
        "omega": omega_pos,
        "phi": phi_pos,
        "kappa": kappa_pos,
        "sam_x": (0.0, 0.0, 0.0),
        "sam_y": (0.0, 0.0, 0.0),
        "sam_z": (0.0, 0.0, 0.0),
    }

    return scan_axis, pos, int(num)


def tristan_writer(
    master_file: Path,
    TR: namedtuple,
    scan_axis: str,
    scan_range: Tuple[float, float],
    timestamps: Tuple[str, str] = (None, None),
):
    """
    A function to call the nexus writer for Tristan 10M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        scan_axis (str): Rotation axis name.
        scan_range (Tuple[float, float]): Start and end value of rotation axis.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to None.
    """
    # Add mask and flatfield file to detector
    detector["pixel_mask"] = maskfile
    detector["flatfield"] = flatfieldfile
    # If these two could instead be passed, I'd be happier...

    # Define OSC scans dictionary
    OSC = {scan_axis: scan_range}

    # Get on with the writing now...
    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))
            #    nxentry.create_dataset("start_time", data=np.string_(timestamps[0]))

            call_writers(
                nxsfile,
                [TR.meta_file],
                coordinate_frame,
                (detector["mode"], None),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                sample_depends_on=scan_axis,
            )

            # write_NXdatetime(nxsfile, (None, timestamps[1]))
            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))
            #    nxentry.create_dataset("end_time", data=np.string_(timestamps[1]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def eiger_writer(
    master_file: Path,
    TR: namedtuple,
    scan_axis: str,
    n_frames: int,
    timestamps: Tuple[str, str] = (None, None),
):
    """
    A function to call the nexus writer for Eiger 2X 4M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        scan_axis (str): Rotation axis name.
        n_frames (int): Number of images.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
    """
    # Find datafiles
    logger.info("Looking for data files ...")
    filename_template = (
        TR.meta_file.parent / TR.meta_file.name.replace("meta", f"{6*'[0-9]'}")
    ).as_posix()
    filenames = [
        Path(f).expanduser().resolve() for f in sorted(glob.glob(filename_template))
    ]
    logger.info(f"Found {len(filenames)} files in directory.")

    # Get scan range array
    logger.info("Calculating scan range...")
    scan_idx = goniometer["axes"].index(scan_axis)

    # Define OSC scans dictionary
    OSC = calculate_scan_range(
        [goniometer["axes"][scan_idx]],
        [goniometer["starts"][scan_idx]],
        [goniometer["ends"][scan_idx]],
        axes_increments=[goniometer["increments"][scan_idx]],
        # n_images=n_frames,
        rotation=True,
    )

    # Get on with the writing now...
    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))

            call_writers(
                nxsfile,
                filenames,
                coordinate_frame,
                (detector["mode"], n_frames),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=None,
                metafile=TR.meta_file,
                link_list=dset_links,
                sample_depends_on=scan_axis,
            )

            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))

            # Write VDS
            image_vds_writer(nxsfile, (int(n_frames), *detector["image_size"]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def write_nxs(**tr_params):
    """
    Gather all parameters from the beamline and call the NeXus writers.
    """
    # Get info from the beamline
    TR = tr_collect(
        meta_file=Path(tr_params["meta_file"]).expanduser().resolve(),
        xml_file=Path(tr_params["xml_file"]).expanduser().resolve(),
        detector_name=tr_params["detector_name"],
        exposure_time=tr_params["exposure_time"],
        wavelength=tr_params["wavelength"],
        beam_center=tr_params["beam_center"],
        start_time=tr_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")  #
        if tr_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=tr_params["stop_time"].strftime(
            "%Y-%m-%dT%H:%M:%S"
        )  # .strftime("%Y-%m-%dT%H:%M:%S")
        if tr_params["stop_time"]
        else None,  # idem.
        geometry_json=tr_params["geometry_json"]
        if tr_params["geometry_json"]
        else None,
        detector_json=tr_params["detector_json"]
        if tr_params["detector_json"]
        else None,
    )

    # Define a file handler
    logfile = TR.meta_file.parent / "nexus_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for beamline I19-2 at DLS.")
    logger.info(f"Detector in use for this experiment: {TR.detector_name}.")
    logger.info(f"Current collection directory: {TR.meta_file.parent}")
    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % TR.meta_file.name)
    # Get NeXus filename
    master_file = get_nexus_filename(TR.meta_file)
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get goniometer and detector parameters
    # FIXME I mean, it works but ...
    if TR.geometry_json:
        logger.info("Reading geometry from json file.")
        _gonio, _det = read_geometry_from_json(TR.geometry_json)
        for k, v in _gonio.items():
            goniometer[k] = v
        for k, v in _det.items():
            detector[k] = v
    else:
        logger.info("Load goniometer from I19-2.")
        for k, v in goniometer_axes.items():
            goniometer[k] = v

    if TR.detector_json:
        logger.info("Reading detector parameters from json file.")
        _det = read_detector_params_from_json(TR.detector_json)
        for k, v in _det.items():
            detector[k] = v
    else:
        logger.info("Load detector parameters for I19-2.")
        if "tristan" in TR.detector_name.lower():
            for k, v in tristan10M_params.items():
                detector[k] = v
        else:
            for k, v in eiger4M_params.items():
                detector[k] = v

    # Read information from xml file
    logger.info("Read xml file.")
    osc_axis, pos, n_frames = read_from_xml(TR.xml_file, TR.detector_name)
    # n_Frames is only useful for eiger
    # pos[scan_axis][::-1] is scan range

    # Finish adding to dictionaries
    # Goniometer
    goniometer["starts"] = []
    goniometer["ends"] = []
    goniometer["increments"] = []
    for ax in goniometer["axes"]:
        goniometer["starts"].append(pos[ax][0])
        goniometer["ends"].append(pos[ax][1])
        goniometer["increments"].append(pos[ax][2])

    # Detector
    detector["exposure_time"] = TR.exposure_time
    detector["beam_center"] = TR.beam_center

    # Module
    if "tristan" in TR.detector_name.lower():
        for k, v in tristan10M_module.items():
            module[k] = v
    else:
        for k, v in eiger4M_module.items():
            module[k] = v

    # Beam
    beam["wavelength"] = TR.wavelength
    beam["flux"] = None

    # Get timestamps in the correct format if they aren't already
    timestamps = (
        get_iso_timestamp(TR.start_time),
        get_iso_timestamp(TR.stop_time),
    )

    logger.info(f"Timestamps recorded: {timestamps}")

    logger.info("Goniometer information")
    logger.info(f"Scan axis is: {osc_axis}")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {TR.beam_center}.")

    if "tristan" in TR.detector_name:
        tristan_writer(master_file, TR, osc_axis, pos[osc_axis][:-1], timestamps)
    else:
        eiger_writer(master_file, TR, osc_axis, n_frames, timestamps)



"""
Command line tool to get an existing .phil file with goniometer/detector metadata or to create a new one.
These files can be used as input for the NeXus generator CLI.
"""

import argparse
import logging
import shutil
import sys

import freephil

try:
    from importlib.resources import files
except ImportError:
    # Python < 3.9 compatibility
    from importlib_resources import files

from pathlib import Path

from .. import log, templates
from . import config_parser, nexus_parser, version_parser

# Define a logger object
logger = logging.getLogger("nexgen.NeXusGenerator")

scopes = freephil.parse(
    """
    include scope nexgen.command_line.nxs_phil.goniometer_scope
    include scope nexgen.command_line.nxs_phil.instrument_scope
    include scope nexgen.command_line.nxs_phil.detector_scope
    include scope nexgen.command_line.nxs_phil.module_scope
    """,
    process_includes=True,
)

parser = argparse.ArgumentParser(description=__doc__, parents=[version_parser])
parser.add_argument("--debug", action="store_const", const=True)


def list_available_phil():
    filedir = files(templates)
    for f in filedir.glob("*.phil"):
        print(f.name)


def get_beamline_phil(args):
    # Determine where to save file
    if args.output:
        odir = Path(args.output).expanduser().resolve()
    else:
        odir = Path(".").expanduser().resolve()
        logger.warning(
            "No output directory was specified by user. A copy of the file will be saved in the current directory."
        )

    # Look for file
    filedir = files(templates)
    found = [f for f in sorted(filedir.glob("*.phil")) if args.phil_file == f.name]
    if len(found) == 0:
        logger.info(f"No {args.phil_file} found.")
    elif len(found) == 1:
        logger.info(f"{args.phil_file} found. Copying in {odir}.")
        shutil.copy(found[0], odir)


def create_new_phil(args):
    cl = scopes.command_line_argument_interpreter()
    working_phil = scopes.fetch(cl.process_and_fetch(args.phil_args))

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    if args.skip is True:
        logger.warning(
            "The scope containing Tristan specs will be removed before writing."
        )
        tristanSpec = [
            obj for obj in working_phil.objects if "tristanSpec" in obj.as_str()
        ][0]
        working_phil.objects.remove(tristanSpec)

    # Write to file
    if args.filename:
        filename = Path(args.filename).expanduser().resolve()
        logger.info(f"Writing new .phil file to {filename.name} in {filename.parent}.")
        with open(filename, "w") as fout:
            fout.write(working_phil.as_str())
    else:
        logger.info(working_phil.as_str())


# Define subparsers
subparser = parser.add_subparsers(
    help="Run nexgen_phil <command> --help to see the options for each command.",
    required=True,
    dest="command",
)

parser_list = subparser.add_parser(
    "list",
    description=("Print out a list of currently available .phil template files."),
)
parser_list.set_defaults(func=list_available_phil)

parser_get = subparser.add_parser(
    "get",
    description=(
        "Get a copy of a .phil file to use as input for the NeXus file writer."
    ),
)
parser_get.add_argument("phil_file", type=str, help="Requested file name.")
parser_get.add_argument("-o", "--output", type=str, help="Specify output directory.")
parser_get.set_defaults(func=get_beamline_phil)

parser_create = subparser.add_parser(
    "new",
    description=("Write a new .phil file."),
    parents=[nexus_parser, config_parser],
)
parser_create.add_argument(
    "-f", "--filename", type=str, help="Filename for new .phil template."
)
parser_create.add_argument(
    "-s",
    "--skip",
    help="Pass to avoid writing tristanSpec scope for non-Tristan detectors.",
    action="store_true",
)
parser_create.set_defaults(func=create_new_phil)


def main():
    # Configure logging
    log.config()

    args = parser.parse_args()
    if args.command == "list":
        args.func()
    else:
        args.func(args)


# main()



"""
Tools to get the information stored inside the _meta.h5 file and overwrite the phil scope.
"""

from __future__ import annotations

import logging
from typing import Any, Dict, List

import h5py

from .. import units_of_length
from .Metafile import DectrisMetafile, TristanMetafile

# TODO actually define the type for scope extract and replace Any with Union
overwrite_logger = logging.getLogger("nexgen.MetaReader")
overwrite_logger.setLevel(logging.DEBUG)


def overwrite_beam(meta_file: h5py.File, name: str, beam: Dict | Any):
    """
    Looks for the wavelength value in the _meta.h5 file.
    If found, it overwrites the value that was parsed from the command line.

    Args:
        meta_file (h5py.File): Handle for _meta.h5 file.
        name (str): Detector description.
        beam (Dict | Any): Scope extract or dictionary defining the beam.

    Raises:
        ValueError: If an invalid detector description is passed. Allowed detectors for this function: Eiger.
    """
    if "eiger" in name.lower():
        meta = DectrisMetafile(meta_file)
        wl = meta.get_wavelength()
        if wl is None:
            overwrite_logger.info("No wavelength information found in meta file.")
            return
    else:
        raise ValueError("Unknown detector: please pass a valid detector description.")
    # If value exists, overwrite. Otherwise, create.
    overwrite_logger.warning("Wavelength will be overwritten.")
    overwrite_logger.info(f"Value for wavelength found in meta file: {wl}")
    if type(beam) is dict:
        beam["wavelength"] = wl
    else:
        try:
            beam.__dict__["wavelength"] = wl
        except KeyError:
            beam.__inject__("wavelength", wl)


def overwrite_detector(
    meta_file: h5py.File, detector: Dict | Any, ignore: List = None
) -> List:
    """
    Looks through the _meta.h5 file for informtion relating to NXdetector.

    Args:
        meta_file (h5py.File): Handle for _meta.h5 file.
        detector (Dict | Any): Scope extract or dictionary defining the detector.
        ignore (List, optional): List of datasets that should not be overwritten by the meta file. Defaults to None.

    Raises:
        ValueError: If an invalid detector description is passed. Allowed detectors for this function: Eiger, Tristan.

    Returns:
        link_list (List): A list of elements to be linked instead of copied in the NeXus file.
    """
    new_values = {}
    link_list = [[], []]
    if type(detector) is dict:
        detector_name = detector["description"].lower()
    else:
        detector_name = detector.description.lower()
    if "tristan" in detector_name:
        meta = TristanMetafile(meta_file)
        new_values["n_modules"] = meta.find_number_of_modules()
        new_values["meta_version"] = meta.find_meta_version()
        link_list[1].append("meta_version")
        overwrite_logger.info("Looking through meta file for Tristan detector.")
        overwrite_logger.info("Number of modules: %d" % new_values["n_modules"])
        overwrite_logger.info(
            "Found meta_version located at: %s " % new_values["meta_version"]
        )
    elif "eiger" in detector_name:
        meta = DectrisMetafile(meta_file)
        overwrite_logger.info("Looking through meta file for Eiger detector.")
        if meta.hasMask is True:
            overwrite_logger.info("Mask has been located in meta file")
            mask_info = meta.find_mask()
            new_values["pixel_mask"] = mask_info[0]
            new_values["pixel_mask_applied"] = mask_info[1]
            link_list[0].append("pixel_mask")
            link_list[0].append("pixel_mask_applied")
        if meta.hasFlatfield is True:
            overwrite_logger.info("Flatfield has been located in meta file")
            flatfield_info = meta.find_flatfield()
            new_values["flatfield"] = flatfield_info[0]
            new_values["flatfield_applied"] = flatfield_info[1]
            link_list[0].append("flatfield")
            link_list[0].append("flatfield_applied")
        if meta.hasDectrisGroup is True:
            new_values["software_version"] = meta.find_software_version()
            new_values["threshold_energy"] = meta.find_threshold_energy()
            new_values["bit_depth_readout"] = meta.find_bit_depth_readout()
            new_values["detector_readout_time"] = meta.find_detector_readout_time()
            new_values["serial_number"] = meta.find_detector_number()
            link_list[0].append("threshold_energy")
            link_list[0].append("bit_depth_readout")
            link_list[0].append("detector_readout_time")
            link_list[0].append("serial_number")
            link_list[1].append("software_version")
            pix = meta.get_pixel_size()
            new_values["pixel_size"] = [
                units_of_length(pix[0]),
                units_of_length(pix[1]),
            ]
            overwrite_logger.warning("Pixel_size will be overwritten.")
            overwrite_logger.info(
                f"Values for x and y pixel size found in meta file: {pix[0]}, {pix[1]}"
            )
            new_values["beam_center"] = meta.get_beam_center()
            overwrite_logger.warning("Beam_center will be overwritten.")
            overwrite_logger.info(
                f"Values for x and y beam center position found in meta file: "
                f"{new_values['beam_center']}"
            )
            sensor_info = meta.get_sensor_information()
            new_values["sensor_material"] = sensor_info[0]
            overwrite_logger.warning("Sensor material will be overwritten.")
            overwrite_logger.info(
                f"Value for sensor material found in meta file: {sensor_info[0]}"
            )
            new_values["sensor_thickness"] = units_of_length(sensor_info[1])
            overwrite_logger.warning("Sensor thickness will be overwritten.")
            overwrite_logger.info(
                f"Value for sensor thickness found in meta file: {sensor_info[1]}"
            )
            new_values["overload"] = meta.get_saturation_value()
            overwrite_logger.warning("Saturation value (overload) will be overwritten.")
            overwrite_logger.info(
                f"Value for overload found in meta file: {new_values['overload']}"
            )
    else:
        overwrite_logger.warning("Unknown detector, exit.")
        raise ValueError("Unknown detector: please pass a valid detector description.")

    if ignore:
        overwrite_logger.warning(
            f"The following datasets are not going to be overwritten: {ignore}"
        )
        for i in ignore:
            if i in new_values:
                del new_values[i]

    for k, v in new_values.items():
        if type(detector) is dict:
            detector[k] = v
        else:
            try:
                detector.__dict__[k] = v
            except KeyError:
                detector.__inject__(k, v)
    return link_list


def update_goniometer(meta_file: h5py.File, goniometer: Dict):
    """
    Read the axes values from the config/ dataset in the meta file and update the goniometer.

    Args:
        meta_file (h5py.File): Handle to Dectris-shaped meta.h5 file.
        goniometer (Dict): Dictionary containing all the goniometer axes information.
    """
    overwrite_logger.info(
        "Get goniometer axes values from meta file for Eiger detector."
    )
    meta = DectrisMetafile(meta_file)

    if meta.hasConfig is True:
        config = meta.read_config_dset()
        num = meta.get_number_of_images()

        goniometer["starts"] = []
        goniometer["ends"] = []
        goniometer["increments"] = []

        for ax in goniometer["axes"]:
            if f"{ax}_start" in config.keys():
                s = config[f"{ax}_start"]
                goniometer["starts"].append(s)
                overwrite_logger.info(f"Start value for axis {ax}: {s}.")
                inc = (
                    config[f"{ax}_increment"]
                    if f"{ax}_increment" in config.keys()
                    else 0.0
                )
                goniometer["increments"].append(inc)
                overwrite_logger.info(f"Increment value for axis {ax}: {inc}.")
                e = s + inc * num
                goniometer["ends"].append(e)
                overwrite_logger.info(f"End value for axis {ax}: {e}.")
            else:
                goniometer["starts"].append(0.0)
                goniometer["ends"].append(0.0)
                goniometer["increments"].append(0.0)
                overwrite_logger.info(f"Axis {ax} not in meta file, values set to 0.0.")
    else:
        overwrite_logger.warning(
            "No config/ dataset found in meta file. Goniometer axes value couldn't be updated from here."
        )
        return


def update_detector_axes(meta_file: h5py.File, detector: Dict):
    """
    Read the axes values from the config/ dataset in the meta file and update the detector.

    Args:
        meta_file (h5py.File): Handle to Dectris-shaped meta.h5 file.
        detector (Dict): Dictionary containing all the detector information.
    """
    overwrite_logger.info("Get detector axes values from meta file for Eiger detector.")
    meta = DectrisMetafile(meta_file)

    if meta.hasConfig is True:
        config = meta.read_config_dset()
        dist = units_of_length(
            meta.get_detector_distance()
        )  # If Dectris file is correct, this is m.

        detector["starts"] = []
        detector["ends"] = []

        # First look for two_theta, then append det_z
        # For the moment, assume the detector doesn't move.
        for ax in detector["axes"]:
            if f"{ax}_start" in config.keys():
                s = config[f"{ax}_start"]
                detector["starts"].append(s)
                overwrite_logger.info(f"Position of axis {ax}: {s}.")

        detector["starts"].append(dist.to("mm").magnitude)
        overwrite_logger.info(f"Position of axis det_z: {dist.to('mm')}.")

        detector["ends"] = detector["starts"]



"""Diamond Light Source beamline utilities."""

from dataclasses import dataclass
from typing import Optional

from dataclasses_json import dataclass_json


@dataclass
@dataclass_json
class PumpProbe:
    """
    Define pump probe parameters.

    Args:
        status (bool): Pump on/off
        exposure (float, optional): Pump exposure time, in s.
        delay (float, optional): Pump delay, in s.
    """

    status: bool = False
    exposure: Optional[float] = None
    delay: Optional[float] = None


source = {
    "name": "Diamond Light Source",
    "short_name": "DLS",
    "type": "Synchrotron X-ray Source",
    "beamline_name": None,
}

eiger_meta_links = [
    [
        "pixel_mask",
        "pixel_mask_applied",
        "flatfield",
        "flatfield_applied",
        "threshold_energy",
        "bit_depth_readout",
        "detector_readout_time",
        "serial_number",
    ],
    ["software_version"],
]



"""
Command line interface to create a NeXus file for time-resolved collections on I19-2.
Available detectors: Tristan 10M, Eiger 2X 4M.
"""

import argparse
import logging
from collections import namedtuple
from datetime import datetime

from .. import log
from . import version_parser

logger = logging.getLogger("nexgen.I19-2_NeXus_cli")

usage = "%(prog)s <sub-command> [options]"
parser = argparse.ArgumentParser(
    usage=usage, description=__doc__, parents=[version_parser]
)
parser.add_argument("--debug", action="store_const", const=True)


def gda_writer(args):
    """
    Write a NeXus file starting from information passed by GDA.
    """
    logger.info("Create a NeXus file for I19-2 interfacing with GDA.")

    from ..beamlines.I19_2_gda_nxs import write_nxs

    write_nxs(
        meta_file=args.meta_file,
        xml_file=args.xml_file,
        detector_name=args.detector_name,  # e.g. "tristan", "eiger",
        exposure_time=float(args.exp_time),
        wavelength=float(args.wavelength),
        beam_center=[
            float(args.beam_center_x),
            float(args.beam_center_y),
        ],
        start_time=datetime.strptime(args.start, "%Y-%m-%dT%H:%M:%SZ")
        if args.start
        else None,
        stop_time=datetime.strptime(args.stop, "%Y-%m-%dT%H:%M:%SZ")
        if args.stop
        else None,  # datetime.now(),
        geometry_json=args.geom if args.geom else None,
        detector_json=args.det if args.det else None,
    )


def nexgen_writer(args):
    """
    Write a NXmx format NeXus file from the I19-2 beamline.
    """
    logger.info("Create a NeXus file for I19-2 data.")

    from ..beamlines.I19_2_nxs import nexus_writer

    if args.axes and not args.ax_start:
        raise IOError(
            "Please pass start and increment values for each of the goniometer axes indicated."
        )
    if args.det_axes and not args.det_start:
        raise IOError(
            "Please pass start and increment values for each of the detector axes indicated."
        )

    if args.axes and args.ax_start:
        if args.detector_name == "eiger":
            axes = namedtuple("axes", ("id", "start", "inc"))
            axes_list = []
            for ax, s, i in zip(args.axes, args.ax_start, args.ax_inc):
                axes_list.append(axes(ax, s, i))
        else:
            axes = namedtuple("axes", ("id", "start", "end"))
            axes_list = []
            for ax, s, e in zip(args.axes, args.ax_start, args.ax_end):
                axes_list.append(axes(ax, s, e))

    if args.det_axes and args.det_start:
        det_axes = namedtuple("det_axes", ("id", "start"))
        det_list = []
        for ax, s in zip(args.det_axes, args.det_start):
            det_list.append(det_axes(ax, s))

    # Check that an actual meta file has been passed and not a data file
    if "meta" not in args.meta_file:
        logger.error(
            f"Wrong input file passed: {args.meta_file}. Please pass the _meta.h5 file for this dataset."
        )
        raise IOError(
            "The input file passed is not a _meta.h5 file. Please pass the correct file."
        )

    nexus_writer(
        meta_file=args.meta_file,
        detector_name=args.detector_name,
        exposure_time=args.exp_time,
        transmission=args.transmission if args.transmission else None,
        wavelength=args.wavelength if args.wavelength else None,
        beam_center=args.beam_center if args.beam_center else None,
        start_time=datetime.strptime(args.start, "%Y-%m-%dT%H:%M:%SZ")
        if args.start
        else None,
        stop_time=datetime.strptime(args.stop, "%Y-%m-%dT%H:%M:%SZ")
        if args.stop
        else None,
        scan_axis=args.scan_axis if args.scan_axis else None,
        gonio_pos=axes_list if args.axes else None,
        det_pos=det_list if args.det_axes else None,
        outdir=args.output if args.output else None,
    )


# Define a couple of useful parsers for axes positions
gonioAx_parser = argparse.ArgumentParser(add_help=False)
gonioAx_parser.add_argument("--axes", type=str, nargs="+", help="Axes names.")
gonioAx_parser.add_argument(
    "--ax-start", type=float, nargs="+", help="Axes start positions."
)
gonioAx_parser.add_argument(
    "--ax-inc", type=float, nargs="+", help="Eventual axes increments."
)
gonioAx_parser.add_argument(
    "--ax-end", type=float, nargs="+", help="Eventual axes ends."
)
gonioAx_parser.add_argument("--scan-axis", type=str, help="Identify scan axis.")

detAx_parser = argparse.ArgumentParser(add_help=False)
detAx_parser.add_argument(
    "--det-axes", type=str, nargs="+", help="Detector axes names."
)
detAx_parser.add_argument(
    "--det-start", type=float, nargs="+", help="Detector axes start positions."
)

# Define subparsers
subparsers = parser.add_subparsers(
    help="Choice depending on how the data collection is run: from GDA or independently of it. \
        Run I19_nexus <command> --help to see the parameters for each sub-command.",
    required=True,
    dest="sub-command",
)

parser_gda = subparsers.add_parser(
    "1",
    aliases=["gda"],
    description=(
        "Trigger the NeXus file writer interface with GDA for I19-2 data collection."
    ),
)
parser_gda.add_argument("meta_file", type=str, help="Path to _meta.h5 file.")
parser_gda.add_argument("xml_file", type=str, help="Path to GDA generated xml file.")
parser_gda.add_argument(
    "detector_name",
    type=str,
    choices=["eiger", "tristan"],
    help="Detector currently in use on beamline.",
)
parser_gda.add_argument("exp_time", type=str, help="Exposure time, in s.")
parser_gda.add_argument("wavelength", type=str, help="Incident beam wavelength.")
parser_gda.add_argument(
    "beam_center_x", type=str, help="Beam center x position, in pixels."
)
parser_gda.add_argument(
    "beam_center_y", type=str, help="Beam center y position, in pixels."
)
parser_gda.add_argument(
    "--start", "--start-time", type=str, default=None, help="Collection start time."
)
parser_gda.add_argument(
    "--stop", "--stop-time", type=str, default=None, help="Collection end time."
)
parser_gda.add_argument(
    "--geom",
    "--geometry-json",
    type=str,
    default=None,
    help="Path to GDA generated geometry json file.",
)
parser_gda.add_argument(
    "--det",
    "--detector-json",
    type=str,
    default=None,
    help="Path to GDA generated detector json file.",
)
parser_gda.set_defaults(func=gda_writer)

parser_nex = subparsers.add_parser(
    "2",
    aliases=["gen"],
    description=("Trigger the NeXus file writer for I19-2 data collection."),
    parents=[gonioAx_parser, detAx_parser],
)
parser_nex.add_argument("meta_file", type=str, help="Path to _meta.h5 file.")
parser_nex.add_argument(
    "detector_name", type=str, help="Detector currently in use on beamline."
)
parser_nex.add_argument("exp_time", type=float, help="Exposure time, in s.")
parser_nex.add_argument(
    "-tr",
    "--transmission",
    type=float,
    default=None,
    help="Attenuator transmission.",
)
parser_nex.add_argument(
    "-wl",
    "--wavelength",
    type=float,
    default=None,
    help="Incident beam wavelength, in A.",
)
parser_nex.add_argument(
    "-bc",
    "--beam-center",
    type=float,
    nargs=2,
    help="Beam center (x,y) positions.",
)
parser_nex.add_argument(
    "--start", "--start-time", type=str, default=None, help="Collection start time."
)
parser_nex.add_argument(
    "--stop", "--stop-time", type=str, default=None, help="Collection end time."
)
parser_nex.add_argument(
    "-o",
    "--output",
    type=str,
    help="Output directory for new NeXus file, if different from collection directory.",
)
parser_nex.set_defaults(func=nexgen_writer)


def main():
    log.config()
    args = parser.parse_args()
    args.func(args)



"""
Writer functions for different groups of a NeXus file.
"""
from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike

from .. import get_iso_timestamp, units_of_length, units_of_time, ureg
from . import (
    calculate_origin,
    create_attributes,
    set_dependency,
    set_instrument_name,
    write_compressed_copy,
)

# from hdf5plugin import Bitshuffle   # noqa: F401


import h5py  # isort: skip


NXclass_logger = logging.getLogger("nexgen.NXclass_writers")
NXclass_logger.setLevel(logging.DEBUG)


# NXentry writer
def write_NXentry(nxsfile: h5py.File, definition: str = "NXmx") -> h5py.Group:
    """
    Write NXentry group at top level of the NeXus file.
    Also, write the application definition NXmx.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        definition (str, optional): Application definition for NeXus file. Defaults to "NXmx".

    Returns:
        nxentry (h5py.Group): NXentry group.
    """
    # Set default attribute
    nxsfile.attrs["default"] = "entry"

    # Start writing the NeXus tree with NXentry at the top level
    nxentry = nxsfile.require_group("entry")
    create_attributes(
        nxentry, ("NX_class", "default", "version"), ("NXentry", "data", "1.0")
    )

    # Application definition: /entry/definition
    nxentry.create_dataset("definition", data=np.string_(definition))
    return nxentry


# NXdata writer
def write_NXdata(
    nxsfile: h5py.File,
    datafiles: List[Path],
    goniometer: Dict,
    data_type: Tuple[str, int],
    osc_scan: Dict[str, ArrayLike],
    transl_scan: Dict[str, ArrayLike] = None,
    entry_key: str = "data",
):
    """
    Write NXdata group at /entry/data.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        datafiles (List[Path]): List of Path objects pointing to HDF5 data files.
        goniometer (Dict): Dictionary containing all the axes information.
        data_type (Tuple[str, int]): Images or events.
        osc_scan (Dict[str, ArrayLike]): Rotation scan. If writing events, this is just a (start, end) tuple.
        transl_scan (Dict[str, ArrayLike], optional): Scan along the xy axes at sample. Defaults to None.
        entry_key (str): Entry key to create the external links to the data files. Defaults to data.

    Raises:
        OSError: If no data is passed.
        ValueError: If the data typs is neither "images" nor "events".
    """
    NXclass_logger.info("Start writing NXdata.")
    # Check that a valid datafile_list has been passed.
    if len(datafiles) == 0:
        raise OSError(
            "No HDF5 data filenames have been found. Please pass at least one."
        )

    # This assumes that a rotation scan is always passed
    osc_axis, osc_range = list(osc_scan.items())[0]

    # Create NXdata group, unless it already exists, in which case just open it.
    nxdata = nxsfile.require_group("/entry/data")
    create_attributes(
        nxdata,
        ("NX_class", "axes", "signal", osc_axis + "_indices"),
        (
            "NXdata",
            osc_axis,
            "data",
            [
                0,
            ],
        ),
    )

    # If mode is images, link to blank image data. Else go to events.
    if data_type[0] == "images":
        tmp_name = f"data_%0{6}d"
        if datafiles[0].parent != Path(nxsfile.filename).parent:
            # This is needed in case the new NeXus file is to be written in a different directory from the data, eg. processing/
            for n, filename in enumerate(datafiles):
                nxdata[tmp_name % (n + 1)] = h5py.ExternalLink(filename, entry_key)
        else:
            for n, filename in enumerate(datafiles):
                nxdata[tmp_name % (n + 1)] = h5py.ExternalLink(filename.name, entry_key)
    elif data_type[0] == "events":
        if len(datafiles) == 1 and "meta" in datafiles[0].as_posix():
            meta = datafiles
        else:
            # Look for meta file to avoid linking to up to 100 files
            tbr = datafiles[0].stem.split("_")[-1]
            mf = datafiles[0].stem.replace(tbr, "meta") + datafiles[0].suffix
            meta = [f for f in datafiles[0].parent.iterdir() if mf in f.as_posix()]
        # If metafile is not found, link to the data files
        if len(meta) == 0:
            for filename in datafiles:
                nxdata[filename.stem] = h5py.ExternalLink(filename.name, "/")
        else:
            nxdata["meta_file"] = h5py.ExternalLink(meta[0].name, "/")
    else:
        raise ValueError(
            "Unknown data type. Please pass one value for data_type from : [images, events]"
        )

    # Write rotation axis dataset
    ax = nxdata.create_dataset(osc_axis, data=osc_range)
    idx = goniometer["axes"].index(osc_axis)
    dep = set_dependency(
        goniometer["depends"][idx], path="/entry/sample/transformations/"
    )

    # Write attributes for axis
    create_attributes(
        ax,
        ("depends_on", "transformation_type", "units", "vector"),
        (
            dep,
            goniometer["types"][idx],
            goniometer["units"][idx],
            goniometer["vectors"][idx],
        ),
    )

    # If present, add linear/grid scan details
    if transl_scan:
        for k, v in transl_scan.items():
            ax_dset = nxdata.create_dataset(k, data=v)
            ax_idx = goniometer["axes"].index(k)
            ax_dep = set_dependency(
                goniometer["depends"][ax_idx], path="/entry/sample/transformations/"
            )
            create_attributes(
                ax_dset,
                ("depends_on", "transformation_type", "units", "vector"),
                (
                    ax_dep,
                    goniometer["types"][ax_idx],
                    goniometer["units"][ax_idx],
                    goniometer["vectors"][ax_idx],
                ),
            )


# NXsample
def write_NXsample(
    nxsfile: h5py.File,
    goniometer: Dict,
    data_type: Tuple[str, int],
    osc_scan: Dict[str, ArrayLike],
    transl_scan: Dict[str, ArrayLike] = None,
    sample_depends_on: str = None,
    sample_details: Dict[str, Any] = None,
):
    """
    Write NXsample group at /entry/sample.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        goniometer (Dict):Dictionary containing all the axes information.
        data_type (Tuple[str, int]): Images or events.
        osc_scan (Dict[str, ArrayLike]): Rotation scan. If writing events, this is just a (start, end) tuple.
        transl_scan (Dict[str, ArrayLike], optional): Scan along the xy axes at sample. Defaults to None.
        sample_depends_on (str, optional): Axis on which the sample depends on. If absent, the depends_on field will be set to the last axis listed in the goniometer. Defaults to None.
        sample_details (Dict[str, Any], optional): General information about the sample, eg. name, temperature.
    """
    NXclass_logger.info("Start writing NXsample and NXtransformations.")
    # Create NXsample group, unless it already exists, in which case just open it.
    nxsample = nxsfile.require_group("/entry/sample")
    create_attributes(
        nxsample,
        ("NX_class",),
        ("NXsample",),
    )

    # Create NXtransformations group: /entry/sample/transformations
    nxtransformations = nxsample.require_group("transformations")
    create_attributes(
        nxtransformations,
        ("NX_class",),
        ("NXtransformations",),
    )

    # Get rotation details
    osc_axis, osc_range = list(osc_scan.items())[0]

    # Save sample depends_on
    if sample_depends_on:
        nxsample.create_dataset(
            "depends_on",
            data=set_dependency(sample_depends_on, path=nxtransformations.name),
        )
    else:
        nxsample.create_dataset(
            "depends_on",
            data=set_dependency(goniometer["axes"][-1], path=nxtransformations.name),
        )

    # Get xy details if passed
    scan_axes = []
    if transl_scan:
        for k in transl_scan.keys():
            scan_axes.append(k)

    # Create sample_{axisname} groups
    for idx, ax in enumerate(goniometer["axes"]):
        grp_name = f"sample_{ax[-1]}" if "sam_" in ax else f"sample_{ax}"
        nxsample_ax = nxsample.create_group(grp_name)
        create_attributes(nxsample_ax, ("NX_class",), ("NXpositioner",))
        if ax == osc_axis:
            # If we're dealing with the scan axis
            if (
                "data" in nxsfile["/entry"].keys()
                and ax in nxsfile["/entry/data"].keys()
            ):
                nxsample_ax[ax] = nxsfile[nxsfile["/entry/data"][ax].name]
                nxtransformations[ax] = nxsfile[nxsfile["/entry/data"][ax].name]
            else:
                nxax = nxsample_ax.create_dataset(ax, data=osc_range)
                _dep = set_dependency(
                    goniometer["depends"][idx], path="/entry/sample/transformations/"
                )
                create_attributes(
                    nxax,
                    ("depends_on", "transformation_type", "units", "vector"),
                    (
                        _dep,
                        goniometer["types"][idx],
                        goniometer["units"][idx],
                        goniometer["vectors"][idx],
                    ),
                )
                nxtransformations[ax] = nxsfile[nxax.name]
            # Write {axisname}_increment_set and {axis_name}_end datasets
            if data_type[0] == "images":
                increment_set = np.repeat(goniometer["increments"][idx], len(osc_range))
                nxsample_ax.create_dataset(
                    ax + "_increment_set",
                    data=goniometer["increments"][idx],
                )  # increment_set
                nxsample_ax.create_dataset(ax + "_end", data=osc_range + increment_set)
        elif ax in scan_axes:
            # For translations
            if (
                "data" in nxsfile["/entry"].keys()
                and ax in nxsfile["/entry/data"].keys()
            ):
                nxsample_ax[ax] = nxsfile[nxsfile["/entry/data"][ax].name]
                nxtransformations[ax] = nxsfile[nxsfile["/entry/data"][ax].name]
            else:
                nxax = nxsample_ax.create_dataset(ax, data=transl_scan[ax])
                _dep = set_dependency(
                    goniometer["depends"][idx], path="/entry/sample/transformations/"
                )
                create_attributes(
                    nxax,
                    ("depends_on", "transformation_type", "units", "vector"),
                    (
                        _dep,
                        goniometer["types"][idx],
                        goniometer["units"][idx],
                        goniometer["vectors"][idx],
                    ),
                )
                nxtransformations[ax] = nxsfile[nxax.name]
        else:
            # For all other axes
            nxax = nxsample_ax.create_dataset(
                ax, data=np.array([goniometer["starts"][idx]])
            )
            _dep = set_dependency(
                goniometer["depends"][idx], path="/entry/sample/transformations/"
            )
            create_attributes(
                nxax,
                ("depends_on", "transformation_type", "units", "vector"),
                (
                    _dep,
                    goniometer["types"][idx],
                    goniometer["units"][idx],
                    goniometer["vectors"][idx],
                ),
            )
            nxtransformations[ax] = nxsfile[nxax.name]

    # Look for nxbeam in file, if it's there make link
    try:
        nxsample["beam"] = nxsfile["/entry/instrument/beam"]
    except KeyError:
        NXclass_logger.debug(
            "No NXbeam group found elsewhere in the NeXus file." "No link written."
        )

    if sample_details:
        for k, v in sample_details.items():
            if type(v) is str:
                v = np.string_(v)
            nxsample.create_dataset(k, data=v)


# NXinstrument
def write_NXinstrument(nxsfile: h5py.File, beam: Dict, attenuator: Dict, source: Dict):
    """
    Write NXinstrument group at /entry/instrument.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        beam (Dict):Dictionary with beam information, mainly wavelength and flux.
        attenuator (Dict): Dictionary containing transmission.
        source (Dict): Dictionary containing the facility information.
    """
    NXclass_logger.info("Start writing NXinstrument.")
    # Create NXinstrument group, unless it already exists, in which case just open it.
    nxinstrument = nxsfile.require_group("/entry/instrument")
    create_attributes(
        nxinstrument,
        ("NX_class",),
        ("NXinstrument",),
    )

    # Write /name field and relative attribute
    NXclass_logger.info(f"{source['short_name']} {source['beamline_name']}")
    name_str = set_instrument_name(source)
    nxinstrument.create_dataset("name", data=np.string_(name_str))
    create_attributes(
        nxinstrument["name"],
        ("short_name",),
        (f"{source['short_name']} {source['beamline_name']}",),
    )

    NXclass_logger.info("Write NXattenuator and NXbeam.")
    # Write NXattenuator group: entry/instrument/attenuator
    nxatt = nxinstrument.require_group("attenuator")
    create_attributes(nxatt, ("NX_class",), ("NXattenuator",))
    if attenuator["transmission"]:
        nxatt.create_dataset(
            "attenuator_transmission",
            data=attenuator["transmission"],
        )

    # Write NXbeam group: entry/instrument/beam
    nxbeam = nxinstrument.require_group("beam")
    create_attributes(nxbeam, ("NX_class",), ("NXbeam",))
    wl = nxbeam.create_dataset("incident_wavelength", data=beam["wavelength"])
    create_attributes(wl, ("units",), ("angstrom",))
    if beam["flux"]:
        flux = nxbeam.create_dataset("total_flux", data=beam["flux"])
        create_attributes(flux, ("units"), ("Hz",))


# NXsource
def write_NXsource(nxsfile: h5py.File, source: Dict):
    """
    Write NXsource group /in entry/source.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        source (Dict): Dictionary containing the facility information.
    """
    NXclass_logger.info("Start writing NXsource.")
    nxsource = nxsfile.require_group("/entry/source")
    create_attributes(
        nxsource,
        ("NX_class",),
        ("NXsource",),
    )

    nxsource.create_dataset("name", data=np.string_(source["name"]))
    create_attributes(nxsource["name"], ("short_name",), (source["short_name"],))
    nxsource.create_dataset("type", data=np.string_(source["type"]))
    if "probe" in source.keys():
        nxsource.create_dataset("probe", data=np.string_(source["probe"]))


# NXdetector writer
def write_NXdetector(
    nxsfile: h5py.File,
    detector: Dict,
    data_type: Tuple[str, int],
    meta: Path = None,
    link_list: List = None,
):
    """
    Write_NXdetector group at /entry/instrument/detector.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        detector (Dict): Dictionary containing all detector information.
        data_type (Tuple[str, int]): Images or events.
        meta (Path, optional): Path to _meta.h5 file. Defaults to None.
        link_list (List, optional): List of values from the meta file to be linked instead of copied. Defaults to None.
    """
    NXclass_logger.info("Start writing NXdetector.")
    # Create NXdetector group, unless it already exists, in which case just open it.
    nxdetector = nxsfile.require_group("/entry/instrument/detector")
    create_attributes(
        nxdetector,
        ("NX_class",),
        ("NXdetector",),
    )

    # Detector description
    nxdetector.create_dataset("description", data=np.string_(detector["description"]))
    nxdetector.create_dataset("type", data=np.string_(detector["detector_type"]))

    # If there is a meta file, a lot of information will be linked instead of copied
    if meta and detector["mode"] == "images":
        NXclass_logger.info(f"Found metadata in {meta.as_posix()} file.")
        meta_link = (
            meta.name
            if meta.parent == Path(nxsfile.filename).parent
            else meta.as_posix()
        )
        for ll in link_list[0]:
            nxdetector[ll] = h5py.ExternalLink(meta_link, detector[ll])
    elif detector["mode"] == "events":
        wd = Path(nxsfile.filename).parent
        # Bad pixel mask
        if detector["pixel_mask"]:
            nxdetector.create_dataset(
                "pixel_mask_applied", data=detector["pixel_mask_applied"]
            )
            NXclass_logger.info(
                f"Looking for file {detector['pixel_mask']} in {wd.as_posix()}."
            )
            maskfile = [
                wd / detector["pixel_mask"]
                for f in wd.iterdir()
                if detector["pixel_mask"] == f.name
            ]
            if maskfile:
                NXclass_logger.info("Pixel mask file found in working directory.")
                write_compressed_copy(
                    nxdetector, "pixel_mask", filename=maskfile[0], dset_key="image"
                )
            else:
                NXclass_logger.warning(
                    "No pixel mask file found in working directory."
                    "Writing and ExternalLink."
                )
                mask = Path(detector["pixel_mask"])
                image_key = (
                    "image" if "tristan" in detector["description"].lower() else "/"
                )
                nxdetector["pixel_mask"] = h5py.ExternalLink(mask.name, image_key)
        # Flatfield
        if detector["flatfield"]:
            nxdetector.create_dataset(
                "flatfield_applied", data=detector["flatfield_applied"]
            )
            NXclass_logger.info(
                f"Looking for file {detector['flatfield']} in {wd.as_posix()}."
            )
            flatfieldfile = [
                wd / detector["flatfield"]
                for f in wd.iterdir()
                if detector["flatfield"] == f.name
            ]
            if flatfieldfile:
                NXclass_logger.info("Flatfield file found in working directory.")
                write_compressed_copy(
                    nxdetector, "flatfield", filename=flatfieldfile[0], dset_key="image"
                )
            else:
                NXclass_logger.warning(
                    "No flatfield file found in the working directory."
                    "Writing and ExternalLink."
                )
                flatfield = Path(detector["flatfield"])
                image_key = (
                    "image" if "tristan" in detector["description"].lower() else "/"
                )
                nxdetector["flatfield"] = h5py.ExternalLink(flatfield.name, image_key)
    else:
        # Flatfield
        if type(detector["flatfield"]) is str:
            nxdetector.create_dataset(
                "flatfield_applied", data=detector["flatfield_applied"]
            )
            flatfield = Path(detector["flatfield"])
            nxdetector["flatfield"] = h5py.ExternalLink(flatfield.name, "/")
        elif detector["flatfield"] is None:
            NXclass_logger.warning(
                "No copy of the flatfield has been found, eithere as a file or dataset."
            )
        else:
            nxdetector.create_dataset(
                "flatfield_applied", data=detector["flatfield_applied"]
            )
            write_compressed_copy(nxdetector, "flatfield", data=detector["flatfield"])
        # Bad pixel mask
        if type(detector["pixel_mask"]) is str:
            nxdetector.create_dataset(
                "pixel_mask_applied", data=detector["pixel_mask_applied"]
            )
            mask = Path(detector["pixel_mask"])
            nxdetector["pixel_mask"] = h5py.ExternalLink(mask.name, "/")
        elif detector["pixel_mask"] is None:
            NXclass_logger.warning(
                "No copy of the pixel_mask has been found, eithere as a file or dataset."
            )
        else:
            nxdetector.create_dataset(
                "pixel_mask_applied", data=detector["pixel_mask_applied"]
            )
            write_compressed_copy(nxdetector, "pixel_mask", data=detector["pixel_mask"])

    # Beam center
    # Check that the information hasn't already been written by the meta file.
    if nxdetector.__contains__("beam_center_x") is False:
        beam_center_x = nxdetector.create_dataset(
            "beam_center_x", data=detector["beam_center"][0]
        )
        create_attributes(beam_center_x, ("units",), ("pixels",))
    if nxdetector.__contains__("beam_center_y") is False:
        beam_center_y = nxdetector.create_dataset(
            "beam_center_y", data=detector["beam_center"][1]
        )
        create_attributes(beam_center_y, ("units",), ("pixels",))

    # Pixel size in m
    if nxdetector.__contains__("x_pixel_size") is False:
        x_pix = units_of_length(detector["pixel_size"][0], True)
        x_pix_size = nxdetector.create_dataset("x_pixel_size", data=x_pix.magnitude)
        create_attributes(x_pix_size, ("units",), (format(x_pix.units, "~"),))
    if nxdetector.__contains__("y_pixel_size") is False:
        y_pix = units_of_length(detector["pixel_size"][1], True)
        y_pix_size = nxdetector.create_dataset("y_pixel_size", data=y_pix.magnitude)
        create_attributes(y_pix_size, ("units",), (format(y_pix.units, "~"),))

    # Sensor material, sensor thickness in m
    if nxdetector.__contains__("sensor_material") is False:
        nxdetector.create_dataset(
            "sensor_material", data=np.string_(detector["sensor_material"])
        )
    if nxdetector.__contains__("sensor_thickness") is False:
        sensor_thickness = units_of_length(detector["sensor_thickness"], True)
        nxdetector.create_dataset("sensor_thickness", data=sensor_thickness.magnitude)
        create_attributes(
            nxdetector["sensor_thickness"],
            ("units",),
            (format(sensor_thickness.units, "~"),),
        )

    # Count time
    if detector["exposure_time"]:
        exp_time = units_of_time(detector["exposure_time"])
        nxdetector.create_dataset("count_time", data=exp_time.magnitude)

    # If detector mode is images write overload and underload
    if (
        data_type[0] == "images"
        and nxdetector.__contains__("saturation_value") is False
    ):
        # if detector["overload"] is not None:
        nxdetector.create_dataset("saturation_value", data=detector["overload"])
        nxdetector.create_dataset("underload_value", data=detector["underload"])

    # Write_NXcollection
    write_NXcollection(nxdetector, detector, data_type, meta, link_list)

    # Write NXtransformations: entry/instrument/detector/transformations/detector_z and two_theta
    nxtransformations = nxdetector.require_group("transformations")
    create_attributes(
        nxtransformations,
        ("NX_class",),
        ("NXtransformations",),
    )

    # Create groups for detector_z and any other detector axis (eg. two_theta) if present
    # This assumes that the detector axes are fixed.
    for idx, ax in enumerate(detector["axes"]):
        if ax == "det_z":
            grp_name = "detector_z"
            dist = units_of_length(str(detector["starts"][idx]) + "mm")  # , True)
        else:
            grp_name = ax

        # It shouldn't be too much of an issue but just in case ...
        if detector["depends"][idx] == "det_z":
            grp_dep = "detector_z"
        else:
            grp_dep = detector["depends"][idx]
        _dep = set_dependency(
            detector["depends"][idx],
            nxtransformations.name + f"/{grp_dep}/",
        )

        nxgrp_ax = nxtransformations.create_group(grp_name)
        create_attributes(nxgrp_ax, ("NX_class",), ("NXpositioner",))
        nxdet_ax = nxgrp_ax.create_dataset(ax, data=np.array([detector["starts"][idx]]))
        create_attributes(
            nxdet_ax,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                _dep,
                detector["types"][idx],
                detector["units"][idx],
                detector["vectors"][idx],
            ),
        )
        if ax == detector["axes"][-1]:
            # Detector depends_on
            nxdetector.create_dataset(
                "depends_on",
                data=set_dependency(ax, path=nxgrp_ax.name),
            )

    # Detector distance
    nxdetector.create_dataset("distance", data=dist.to("m").magnitude)
    create_attributes(
        nxdetector["distance"], ("units",), (format(dist.to("m").units, "~"))
    )

    # Check if there are any remaining datasets to be written (usually from the meta file but not always)
    others = [
        "threshold_energy",
        "bit_depth_readout",
        "detector_number",
        "detector_readout_time",
        "photon_energy",
    ]
    for dset in others:
        if nxdetector.__contains__(dset) is False and dset in detector.keys():
            val = (
                np.string_(detector[dset])
                if type(detector[dset]) is str
                else detector[dset]
            )
            if val is not None:  # FIXME bit of a gorilla here, for bit_depth_readout
                nxdetector.create_dataset(dset, data=val)


# NXdetector_module writer
def write_NXdetector_module(
    nxsfile: h5py.File,
    module: Dict,
    image_size: List | Tuple,
    pixel_size: List | Tuple,
    beam_center: Optional[List | Tuple] = None,
):
    """
    Write NXdetector_module group at /entry/instrument/detector/module.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        module (Dict): Dictionary containing the detector module information: fast and slow axes, how many modules.
        image_size (List | Tuple): Size of the detector.
        pixel_size (List | Tuple): Size of the single pixels in fast and slow direction, in mm.
        beam_center (Optional[List | Tuple], optional): Beam center position, needed only if origin needs to be calculated. Defaults to None.
    """
    NXclass_logger.info("Start writing NXdetector_module.")
    # Create NXdetector_module group, unless it already exists, in which case just open it.
    nxmodule = nxsfile.require_group("/entry/instrument/detector/module")
    create_attributes(
        nxmodule,
        ("NX_class",),
        ("NXdetector_module",),
    )

    # TODO check how many modules, and write as many, plus NXdetector_group
    nxmodule.create_dataset("data_origin", data=np.array([0, 0]))
    nxmodule.create_dataset("data_size", data=image_size)
    nxmodule.create_dataset("data_stride", data=np.array([1, 1]))

    # Write fast_ and slow_ pixel_direction
    fast_axis = tuple(module["fast_axis"])
    slow_axis = tuple(module["slow_axis"])

    if "offsets" in module.keys():
        offsets = module["offsets"]
    else:
        offsets = [(0, 0, 0), (0, 0, 0)]

    x_pix = units_of_length(pixel_size[0], True)
    fast_pixel = nxmodule.create_dataset("fast_pixel_direction", data=x_pix.magnitude)
    create_attributes(
        fast_pixel,
        (
            "depends_on",
            "offset",
            "offset_units",
            "transformation_type",
            "units",
            "vector",
        ),
        (
            "/entry/instrument/detector/transformations/detector_z/det_z",
            offsets[0],
            "mm",
            "translation",
            format(x_pix.units, "~"),
            fast_axis,
        ),
    )

    y_pix = units_of_length(pixel_size[1], True)
    slow_pixel = nxmodule.create_dataset("slow_pixel_direction", data=y_pix.magnitude)
    create_attributes(
        slow_pixel,
        (
            "depends_on",
            "offset",
            "offset_units",
            "transformation_type",
            "units",
            "vector",
        ),
        (
            "/entry/instrument/detector/module/fast_pixel_direction",
            offsets[1],
            "mm",
            "translation",
            format(y_pix.units, "~"),
            slow_axis,
        ),
    )

    # If module_offset is set to 1 or 2, calculate accordinlgy and write the field
    if "module_offset" not in module.keys():
        NXclass_logger.warning(
            "Module_offset option wasn't passed."
            "It will be automatically set to '1' and the origin calculated accordingly."
            "To skip this calculation please set module_offset to '0'."
        )
        module["module_offset"] = "1"

    if module["module_offset"] != "0":
        pixel_size_m = [
            x_pix.magnitude,
            y_pix.magnitude,
        ]
        origin, offset_val = calculate_origin(
            beam_center,
            pixel_size_m,
            fast_axis,
            slow_axis,
            mode=module["module_offset"],
        )
        module_offset = nxmodule.create_dataset("module_offset", data=offset_val)
        create_attributes(
            module_offset,
            (
                "depends_on",
                "offset",
                "offset_units",
                "transformation_type",
                "units",
                "vector",
            ),
            (
                "/entry/instrument/detector/transformations/detector_z/det_z",
                [0, 0, 0],
                "mm",
                "translation",
                format(x_pix.units, "~"),
                origin,
            ),
        )

        # Correct dependency tree accordingly
        _path = "/entry/instrument/detector/module/module_offset"
        create_attributes(fast_pixel, ("depends_on",), (_path,))
        create_attributes(slow_pixel, ("depends_on",), (_path,))


# NXdetector_group writer
# def writr_NXdetector_group(nxinstrument: h5py.Group, detector: Dict):
# TODO to be added once multiple module functionality works
# pass


# NXCollection writer (detectorSpecific)
def write_NXcollection(
    nxdetector: h5py.Group,
    detector: Dict,
    data_type: Tuple[str, int],
    meta: Path = None,
    link_list: List = None,
):
    """
    Write a NXcollection group inside NXdetector as detectorSpecific.

    Args:
        nxdetector (h5py.Group): HDF5 NXdetector group handle.
        detector (Dict): Dictionary containing all detector information.
        data_type (Tuple[str, int]): Images or events.
        meta (Path, optional): Path to _meta.h5 file. Defaults to None.
        link_list (List, optional): List of values from the meta file to be linked instead of copied. Defaults to None.
    """
    NXclass_logger.info("Start writing detectorSpecific group as NXcollection.")
    # Create detectorSpecific group
    grp = nxdetector.require_group("detectorSpecific")
    grp.create_dataset("x_pixels", data=detector["image_size"][1])  # fast axis
    grp.create_dataset("y_pixels", data=detector["image_size"][0])  # slow axis
    if data_type[0] == "images":
        grp.create_dataset("nimages", data=data_type[1])
    if meta and data_type[0] == "images":
        for l in link_list[1]:
            grp[l] = h5py.ExternalLink(meta.name, detector[l])
    else:
        if "software_version" in detector:
            grp.create_dataset(
                "software_version", data=np.string_(detector["software_version"])
            )
    if "TRISTAN" in detector["description"].upper():  # or data_type[1] == "events":
        tick = ureg.Quantity(detector["detector_tick"])
        grp.create_dataset("detector_tick", data=tick.magnitude)
        grp["detector_tick"].attrs["units"] = np.string_(format(tick.units, "~"))
        freq = ureg.Quantity(detector["detector_frequency"])
        grp.create_dataset("detector_frequency", data=freq.magnitude)
        grp["detector_frequency"].attrs["units"] = np.string_(format(freq.units, "~"))
        grp.create_dataset(
            "timeslice_rollover_bits", data=detector["timeslice_rollover"]
        )


# NXdatetime writer
def write_NXdatetime(nxsfile: h5py.File, timestamps: List | Tuple):
    """
    Write start and end timestamps under /entry/start_time and /entry/end_time.

    Args:
        nxsfile (h5py.File): Nexus file to be written.
        timestamps (List | Tuple): Timestamps (start, end).
    """
    nxentry = nxsfile.require_group("entry")

    start = timestamps[0]
    if start:
        if type(start) is datetime:
            start = start.strftime("%Y-%m-%dT%H:%M:%S")
            start = get_iso_timestamp(start)
        if start.endswith("Z") is False:  # Just in case
            start += "Z"
        nxentry.create_dataset("start_time", data=np.string_(start))

    stop = timestamps[1]
    if stop:
        if type(stop) is datetime:
            stop = stop.strftime("%Y-%m-%dT%H:%M:%S")
            stop = get_iso_timestamp(stop)
        if stop.endswith("Z") is False:
            stop += "Z"
        nxentry.create_dataset("end_time", data=np.string_(stop))


# NXnote writer
# To be used e.g. as a place to store pump-probe info such as pump delay/width
def write_NXnote(nxsfile: h5py.File, loc: str, info: Dict):
    """
    Write any additional information as a NXnote class in a specified location in the NeXus file.

    Args:
        nxsfile (h5py.File): NeXus file handle.
        loc (str): Location inside the NeXus file to write NXnote group.
        info (Dict): Dictionary of datasets to be written to NXnote.
    """
    NXclass_logger.info("Start writing NXnote.")
    # Create the NXnote group in the specified location
    nxnote = nxsfile.require_group(loc)
    create_attributes(
        nxnote,
        ("NX_class",),
        ("NXnote",),
    )

    # Write datasets
    for k, v in info.items():
        if v:  # Just in case one value is not recorded and set as None
            if type(v) is str:
                v = np.string_(v)
            nxnote.create_dataset(k, data=v)
            NXclass_logger.info(f"{k} dataset written in {loc}.")


def write_NXcoordinate_system_set(
    nxsfile: h5py.File,
    convention: str,
    base_vectors: Dict[str, Tuple],
    origin: List | Tuple | ArrayLike,
):
    """
    Write a container object to store coordinate system conventions different from mcstas.

    The NXcoordinate_system_set base class is used here to define and store the mappings between mcstas and a different coordinate system relevant to the data.
    It should hold at least one NXtransformations group containing a depends_on field, which specifies whether this coordinate system is the reference ("."),
    as well as the three base vectors and the location of the origin.

    A template for base_vectors:
    base_vectors = {"x": (depends_on, transformation_type, units, vector), ...}

    Args:
        nxsfile (h5py.File): Handle to NeXus file.
        convention (str): Convention decription. Defaults to "ED".
        base_vectors (Dict[str, Tuple]): The three base vectors of the coordinate system.
        origin (List | Tuple | np.ndarray): The location of the origin of the coordinate system.
    """
    NXclass_logger.info(
        f"Writing NXcoordinate_system_set to define the coordinate system convention for {convention}."
    )

    #
    nxcoord = nxsfile.require_group("/entry/coordinate_system_set")
    create_attributes(
        nxcoord,
        ("NXclass",),
        ("NXcoordinate_system_set",),
    )

    transf = nxcoord.require_group("transformations")
    create_attributes(
        transf,
        ("NXclass",),
        ("NXtransformations",),
    )

    # Needs at least: 3 base vectors, depends_on ("." probably?), origin
    transf.create_dataset("depends_on", data=np.string_("."))  # To be checked
    transf.create_dataset("origin", data=origin)

    # Base vectors
    NXclass_logger.info(
        "Base vectors: \n"
        f"x: {base_vectors['x'][-1]} \n"
        f"y: {base_vectors['y'][-1]} \n"
        f"z: {base_vectors['z'][-1]} \n"
    )
    idx = 0
    for k, v in base_vectors.items():
        base = transf.create_dataset(k, data=np.array(origin[idx]))
        create_attributes(
            base,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                set_dependency(v[0], transf.name),
                v[1],
                v[2],
                v[3],
            ),
        )
        idx += 1



"""
Create a NeXus file for serial crystallography datasets collected on Eiger detector either on I19-2 or I24 beamlines.
"""
from __future__ import annotations

import glob
import logging
from collections import namedtuple
from pathlib import Path

import h5py

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write.NexusWriter import call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry, write_NXnote
from ..tools.MetaReader import update_detector_axes, update_goniometer
from ..tools.VDS_tools import image_vds_writer
from . import PumpProbe, eiger_meta_links, source

__all__ = ["ssx_eiger_writer"]

# Define logger
logger = logging.getLogger("nexgen.SSX_Eiger")

# Define a namedtuple for collection parameters
ssx_collect = namedtuple(
    "ssx_collect",
    [
        "num_imgs",
        "exposure_time",
        "detector_distance",
        "beam_center",
        "transmission",
        "wavelength",
        "start_time",
        "stop_time",
        "chip_info",
        "chipmap",
    ],
)

ssx_collect.__doc__ = """Serial collection parameters"""

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize some dictionaries
module = {}
beam = {}
attenuator = {}


def ssx_eiger_writer(
    visitpath: Path | str,
    filename: str,
    beamline: str,
    expt_type: str = "fixed-target",
    pump_status: bool = False,
    **ssx_params,
):
    """Gather all collection parameters and write the NeXus file for SSX using Eiger detector.

    Args:
        visitpath (Path | str): Collection directory.
        filename (str): Filename root.
        beamline (str): Beamline on which the experiment is being run.
        expt_type (str, optional): Experiment type, accepted values: extruder,
            fixed-target, 3Dgridscan. Defaults to "fixed-target".
        pump_status (bool, optional): True for pump-probe experiment. Defaults to False.

    Keyword Args:
        num_imgs (int): Total number of images collected.
        exp_time (float): Exposure time, in s.
        det_dist (float): Distance between sample and detector, in mm.
        beam_center (List[float, float]): Beam center position, in pixels.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        flux (float): Total flux.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        chip_info (Dict): For a grid scan, dictionary containing basic chip information.
            At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures.
        chipmap (Path | str): Path to the chipmap file corresponding to the experiment,
            or 'fullchip' indicating that the whole chip is being scanned.
        pump_exp (float): Pump exposure time, in s.
        pump_delay (float): Pump delay time, in s.
        osc_axis (str): Oscillation axis. Always omega on I24. If not passed it will default to phi for I19-2.

    Raises:
        ValueError: If an invalid beamline name is passed.
        ValueError: If an invalid experiment type is passed.
    """
    SSX = ssx_collect(
        num_imgs=int(ssx_params["num_imgs"]),
        exposure_time=ssx_params["exp_time"],
        detector_distance=ssx_params["det_dist"]
        if "det_dist" in ssx_params.keys()
        else None,
        beam_center=ssx_params["beam_center"],
        transmission=ssx_params["transmission"]
        if "transmission" in ssx_params.keys()
        else None,
        wavelength=ssx_params["wavelength"]
        if "wavelength" in ssx_params.keys()
        else None,
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,
        chip_info=ssx_params["chip_info"] if "chip_info" in ssx_params.keys() else None,
        chipmap=ssx_params["chipmap"] if "chipmap" in ssx_params.keys() else None,
    )

    visitpath = Path(visitpath).expanduser().resolve()

    # Configure logging
    logfile = visitpath / f"{beamline}_EigerSSX_nxs_writer.log"
    log.config(logfile.as_posix())

    logger.info(f"Current collection directory: {visitpath}")
    # Find metafile in directory and get info from it
    try:
        metafile = [
            f for f in visitpath.iterdir() if filename + "_meta" in f.as_posix()
        ][0]
        logger.info(f"Found {metafile} in directory.")
    except IndexError as err:
        logger.exception(err)
        logger.error(
            "Missing metadata, something might be wrong with this collection."
            "Unable to write NeXus file at this time. Please try using command line tool."
        )
        raise

    # Find datafiles
    filename_template = (
        metafile.parent / metafile.name.replace("meta", f"{6*'[0-9]'}")
    ).as_posix()
    # if meta else (SSX.visitpath / SSX.filename).as_posix() + f"_{6*'[0-9]'}.h5"
    filename = [
        Path(f).expanduser().resolve() for f in sorted(glob.glob(filename_template))
    ]
    logger.info(f"Found {len(filename)} data files in directory.")

    logger.info("Creating a NeXus file for %s ..." % metafile.name)
    # Get NeXus filename
    master_file = get_nexus_filename(filename[0])
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get parameters depending on beamline
    logger.info(f"DLS Beamline: {beamline.upper()}.")
    if "I19" in beamline.upper():
        osc_axis = ssx_params["osc_axis"] if "osc_axis" in ssx_params.keys() else "phi"
        from .I19_2_params import eiger4M_module as module
        from .I19_2_params import eiger4M_params as detector
        from .I19_2_params import goniometer_axes as goniometer

        source["beamline_name"] = beamline.upper()
        beam["flux"] = None
    elif "I24" in beamline.upper():
        osc_axis = "omega"
        from .I24_Eiger_params import eiger9M_module as module
        from .I24_Eiger_params import eiger9M_params as detector
        from .I24_Eiger_params import goniometer_axes as goniometer

        source["beamline_name"] = beamline.upper()
        beam["flux"] = ssx_params["flux"] if "flux" in ssx_params.keys() else None
    else:
        raise ValueError(
            "Unknown beamline for SSX collections with Eiger detector."
            "Beamlines currently enabled for the writer: I24 (Eiger 9M), I19-2 (Eiger 4M)."
        )

    # Add to dictionaries
    attenuator["transmission"] = SSX.transmission

    detector["exposure_time"] = SSX.exposure_time
    detector["beam_center"] = SSX.beam_center

    # Look for wavelength
    if SSX.wavelength:
        beam["wavelength"] = SSX.wavelength
    else:
        logger.debug("No wavelength passed, looking for it in the meta file.")
        from ..tools.Metafile import DectrisMetafile

        with h5py.File(metafile, "r") as fh:
            _wl = DectrisMetafile(fh).get_wavelength()
            beam["wavelength"] = _wl

    # Look for detector distance
    with h5py.File(metafile, "r") as fh:
        update_goniometer(fh, goniometer)
        update_detector_axes(fh, detector)
    logger.debug(
        "Goniometer and detector axes have ben updated with values from the meta file."
    )
    det_z_idx = detector["axes"].index("det_z")
    if SSX.detector_distance and SSX.detector_distance != detector["starts"][det_z_idx]:
        logger.debug(
            "Detector distance value in meta file did not match with the one passed by the user.\n"
            f"Passed value: {SSX.detector_distance}; Value stored in meta file: {detector['starts'][det_z_idx]}.\n"
            "Value will be overwritten with the passed one."
        )
        detector["starts"][det_z_idx] = SSX.detector_distance

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX.start_time),
        get_iso_timestamp(SSX.stop_time),
    )

    # Get pump information
    pump_probe = PumpProbe()
    if pump_status is True:
        # Exposure and delay could also be found in dictionary for grid scan
        logger.info("Pump status is True.")
        pump_probe.status = pump_status
        pump_probe.exposure = (
            ssx_params["pump_exp"] if "pump_exp" in ssx_params.keys() else None
        )
        pump_probe.delay = (
            ssx_params["pump_exp"] if "pump_exp" in ssx_params.keys() else None
        )

        logger.info(f"Recorded pump exposure time: {pump_probe.exposure}")
        logger.info(f"Recorded pump delay time: {pump_probe.delay}")

    # Define what to do based on experiment type
    if expt_type not in ["extruder", "fixed-target", "3Dgridscan"]:
        raise ValueError(
            "Please pass a valid experiment type.\n"
            "Accepted values: extruder, fixed-target, 3Dgridscan."
        )

    if expt_type == "extruder":
        from .SSX_expt import run_extruder

        goniometer, OSC, pump_info = run_extruder(goniometer, SSX.num_imgs, pump_probe)
        TRANSL = None
        tot_num_imgs = SSX.num_imgs
    elif expt_type == "fixed-target":
        # Define chipmap if needed
        chipmapfile = (
            "fullchip"
            if SSX.chipmap is None
            else Path(SSX.chipmap).expanduser().resolve()
        )
        from .SSX_expt import run_fixed_target

        # I19-2 meta file sanity check
        logger.debug(
            "Sanity check. There is no rotation here.\n"
            "Setting all rotation values to same start and end for this application."
        )
        goniometer["ends"] = [s for s in goniometer["starts"]]
        logger.debug(f"Starts: {goniometer['starts']}. Ends: {goniometer['ends']}")
        goniometer, OSC, TRANSL, pump_info = run_fixed_target(
            goniometer, SSX.chip_info, chipmapfile, pump_probe, osc_axis=osc_axis
        )
        # Check that things make sense
        if SSX.num_imgs != len(TRANSL["sam_x"]):
            logger.warning(
                f"The total number of scan points is {len(TRANSL['sam_x'])}, which does not match the total number of images passed as input {SSX.num_imgs}."
            )
            logger.warning(
                "Reset SSX.num_imgs to number of scan points for vds creation."
            )
            tot_num_imgs = len(TRANSL["sam_x"])
        else:
            tot_num_imgs = SSX.num_imgs
        tot_num_imgs = len(TRANSL["sam_x"])
    elif expt_type == "3Dgridscan":
        # Define chipmap if needed
        chipmapfile = (
            "fullchip"
            if SSX.chipmap is None
            else Path(SSX.chipmap).expanduser().resolve()
        )
        from .SSX_expt import run_3D_grid_scan

        goniometer, OSC, TRANSL, pump_info = run_3D_grid_scan(
            goniometer, SSX.chip_info, chipmapfile, pump_probe, osc_axis=osc_axis
        )
        tot_num_imgs = len(TRANSL["sam_x"])

    logger.info("--- COLLECTION SUMMARY ---")
    logger.info("Source information")
    logger.info(f"Facility: {source['name']} - {source['type']}.")
    logger.info(f"Beamline: {source['beamline_name']}")

    logger.info(f"Incident beam wavelength: {beam['wavelength']}")
    logger.info(f"Attenuation: {attenuator['transmission']}")

    logger.info("Goniometer information")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info(f"Oscillation axis: {list(OSC.keys())[0]}.")
    if expt_type != "extruder":
        logger.info(f"Grid scan axes: {list(TRANSL.keys())}.")

    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {detector['beam_center']}.")

    logger.info(f"Timestamps recorded: {timestamps}")

    # Get to the actual writing
    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            call_writers(
                nxsfile,
                filename,
                coordinate_frame,
                (detector["mode"], tot_num_imgs),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=TRANSL,
                metafile=metafile,
                link_list=eiger_meta_links,
            )

            if pump_status is True:
                logger.info("Write pump information to file.")
                loc = "/entry/source/notes"
                write_NXnote(nxsfile, loc, pump_info)

            # Write VDS
            # TODO discuss how VDS should be saved. All in one probably not ideal for N_EXPOSURES > 1.
            image_vds_writer(nxsfile, (int(tot_num_imgs), *detector["image_size"]))

            if timestamps:
                write_NXdatetime(nxsfile, timestamps)

            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise



"""
General tools useful to create NeXus format files.
"""
from __future__ import annotations

__author__ = "Diamond Light Source - Scientific Software"
__email__ = "data_analysis@diamond.ac.uk"
__version__ = "0.6.21"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))

import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Tuple

import h5py
import numpy as np
import pint
from numpy.typing import ArrayLike

# Logging set up
logging.getLogger("nexgen").addHandler(logging.NullHandler())

# Initialize registry and a Quantity constructor
ureg = pint.UnitRegistry()
Q_ = ureg.Quantity

# Define scope extract type
# Scope = freephil.common.scope_extract

# Filename pattern: filename_######.h5 or filename_meta.h5
# P = re.compile(r"(.*)_(?:\d+)")
P = re.compile(r"(.*)_(?:meta|\d+)")

# Format strings for timestamps
format_list = [
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%d %H:%M:%S",
    "%a %b %d %Y %H:%M:%S",
    "%A, %d. %B %Y %I:%M%p",
]


def imgcif2mcstas(vector: List | Tuple | ArrayLike) -> Tuple:
    """
    Convert from the standard coordinate frame used by imgCIF/CBF to the
    NeXus McStas coordinate system.

    Args:
        vector (List | Tuple | np.array): Coordinates to be converted.

    Returns:
        Tuple: Converted coordinate values.
    """
    c2n = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])
    return tuple(np.dot(c2n, vector))


def coord2mcstas(vector: List | Tuple | ArrayLike, mat: ArrayLike) -> Tuple:
    """
    General conversion from a new coordinate convention to the NeXus McStas coordinate system.

    Args:
        vector (List | Tuple | np.array): Coordinates to be converted.
        mat (np.ndarray): Coordinate transformation matrix.

    Returns:
        Tuple: Converted coordinate values
    """
    return tuple(np.dot(mat, vector))


def get_filename_template(input_filename: Path) -> str:
    """
    Get the data file name template from either the master or the meta file.

    Args:
        input_filename (Path): Path object containing the name of master or meta file.
                            The format should be either file_master.h5, file.nxs for a master file, file_meta.h5 for a meta file.

    Raises:
        NameError: If the input file does not have the expected format.

    Returns:
        filename_template (str): String template for the name of blank data file.
    """
    if input_filename.suffix == ".nxs":
        filename_root = input_filename.stem
        filename_template = input_filename.parent / f"{filename_root}_%0{6}d.h5"
    elif input_filename.suffix == ".h5" and "master" in input_filename.as_posix():
        filename = input_filename.stem.replace("master", f"%0{6}d")
        filename_template = input_filename.parent / f"{filename}.h5"
    elif input_filename.suffix == ".h5" and "meta" in input_filename.as_posix():
        filename = input_filename.stem.replace("meta", f"%0{6}d")
        filename_template = input_filename.parent / f"{filename}.h5"
    else:
        raise NameError(
            "Input file did not have the expected format for a master or meta file."
        )
    # so that filename_template.as_posix() % 1 will become filename_000001.h5
    return filename_template.as_posix()


def get_nexus_filename(input_filename: Path, copy: bool = False) -> Path:
    """
    Get the filename for the NeXus file from the stem of the input file name.

    Args:
        input_filename (Path): File name and path of either a .h5 data file or a _meta.h5 file.
        copy (bool, optional): Avoid trying to write a new file with the same name as the old one when making a copy. Defaults to False.

    Returns:
        Path: NeXus file name (.nxs) path.
    """
    filename_stem = P.fullmatch(input_filename.stem)
    if filename_stem:
        filename = filename_stem[1]
    else:
        filename = input_filename.stem

    if copy is True:
        nxs_filename = input_filename.parent / f"{filename}_copy.nxs"
    else:
        nxs_filename = input_filename.parent / f"{filename}.nxs"
    return nxs_filename


def walk_nxs(nxs_obj: h5py.File | h5py.Group) -> List[str]:
    """
    Walk all the groups, subgroups and datasets of an object.

    Args:
        nxs_obj (h5py.File | h5py.Group): Object to walk through, could be a file or a group.

    Returns:
        obj_list (List[str]): List of objects found, as strings.
    """
    obj_list = []
    nxs_obj.visit(obj_list.append)
    return obj_list


def split_arrays(axes_names: List, array: List) -> Dict[str, Tuple]:
    """Split a list of values into arrays.

    This function splits up the list of values passed as input (eg. phil parameters, dictionary) \
    for vector, offset for all existing axes.

    Args:
        axes_names (List): Axes names.
        array (List): Array of values to be split up. It must be

    Raises:
        ValueError: When each axes doesn't have a corresponding array of size 3.

    Returns:
        array_dict (Dict[str, Tuple]): Dictionary of arrays corresponding to each axis. Keys are axes names.
    """
    array_dict = {}
    if len(axes_names) == len(array):
        array_dict = {ax: tuple(v) for ax, v in zip(axes_names, array)}
        return array_dict
    elif len(array) == 3 * len(axes_names):
        for j in range(len(axes_names)):
            a = array[3 * j : 3 * j + 3]
            array_dict[axes_names[j]] = tuple(a)
        return array_dict
    else:
        raise ValueError(
            f"Number of axes {len(axes_names)} doesn't match the lenght of the array list {len(array)}."
            "Please check again and make sure that all axes have a matching array of size 3."
        )


def reframe_arrays(
    goniometer: Dict[str, Any],
    detector: Dict[str, Any],
    module: Dict[str, Any],
    coordinate_frame: str = "mcstas",
    new_coord_system: Dict[str, Any] = None,
):
    """
    Split a list of offset/vector values into arrays. If the coordinate frame is not mcstas, \
    convert the arrays using the base vectors of the new coordinate system.

    Args:
        goniometer (Dict[str, Any]): Goniometer geometry description.
        detector (Dict[str, Any]): Detector specific parameters and its axes.
        module (Dict[str, Any]): Geometry and description of detector module.
        coordinate_frame (str, optional): Coordinate system being used. If "imgcif", there's no need to pass a \
            new coordinate system definition, as the conversion is already included in nexgen. Defaults to "mcstas".
        new_coord_system (Dict[str, Any], optional): Definition of the current coordinate system. \
            It should at least contain a string defining the convention, origin and axes information as a tuple of (depends_on, type, units, vector). \
            e.g. for X axis: {"x": (".", "translation", "mm", [1,0,0])}. \
            Defaults to None.

    Raises:
        ValueError: When the input coordinate system name and the coordinate system convention for the vectors doesn't match.
    """
    # If the arrays of vectors/offsets are not yet split, start by doing that
    goniometer["vectors"] = list(
        split_arrays(goniometer["axes"], goniometer["vectors"]).values()
    )
    goniometer["offsets"] = list(
        split_arrays(goniometer["axes"], goniometer["offsets"]).values()
    )

    detector["vectors"] = list(
        split_arrays(detector["axes"], detector["vectors"]).values()
    )

    if "offsets" in module.keys():
        module["offsets"] = list(
            split_arrays(["fast_axis", "slow_axis"], module["offsets"]).values()
        )

    # Now proceed with conversion if needed
    if coordinate_frame.lower() != "mcstas":
        if coordinate_frame.lower() == "imgcif":
            # Goniometer
            goniometer["vectors"] = [imgcif2mcstas(v) for v in goniometer["vectors"]]
            goniometer["offsets"] = [imgcif2mcstas(v) for v in goniometer["offsets"]]

            # Detector
            detector["vectors"] = [imgcif2mcstas(v) for v in detector["vectors"]]

            # Module
            module["fast_axis"] = imgcif2mcstas(module["fast_axis"])
            module["slow_axis"] = imgcif2mcstas(module["slow_axis"])
            if "offsets" in module.keys():
                module["offsets"] = [imgcif2mcstas(off) for off in module["offsets"]]
        else:
            if coordinate_frame != new_coord_system["convention"]:
                raise ValueError(
                    "The input coordinate frame value doesn't match the current cordinate system convention."
                    "Impossible to convert to mcstas."
                )
            mat = np.array(
                [
                    new_coord_system["x"][-1],
                    new_coord_system["y"][-1],
                    new_coord_system["z"][-1],
                ]
            )

            # Goniometer
            goniometer["vectors"] = [
                coord2mcstas(v, mat) for v in goniometer["vectors"]
            ]
            goniometer["offsets"] = [
                coord2mcstas(v, mat) for v in goniometer["offsets"]
            ]

            # Detector
            detector["vectors"] = [coord2mcstas(v, mat) for v in detector["vectors"]]

            # Module
            module["fast_axis"] = coord2mcstas(module["fast_axis"], mat)
            module["slow_axis"] = coord2mcstas(module["slow_axis"], mat)
            if "offsets" in module.keys():
                module["offsets"] = [
                    coord2mcstas(off, mat) for off in module["offsets"]
                ]


def get_iso_timestamp(ts: str | float) -> str:
    """
    Format a timestamp string to be stores in a NeXus file according to ISO8601: 'YY-MM-DDThh:mm:ssZ'

    Args:
        ts (str | float): Input string, can also be a timestamp (eg. time.time()) string.
                        Allowed formats: "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S", "%a %b %d %Y %H:%M:%S", "%A, %d. %B %Y %I:%M%p".

    Returns:
        ts_iso (str): Formatted timestamp.
    """
    if ts is None:
        return None
    try:
        ts = float(ts)
        ts_iso = datetime.utcfromtimestamp(ts).replace(microsecond=0).isoformat()
    except ValueError:
        for fmt in format_list:
            try:
                ts_iso = datetime.strptime(ts, fmt).isoformat()
            except ValueError:
                ts_iso = str(ts)
    if ts_iso.endswith("Z") is False:
        ts_iso += "Z"
    return ts_iso


def units_of_length(q: Any, to_base: bool = False) -> Q_:  # -> pint.Quantity:
    """
    Check that a quantity of length is compatible with NX_LENGTH, defaulting to m if dimensionless.

    Args:
        q (Any): An object that can be interpreted as a pint Quantity, it can be dimensionless.
        to_base (bool, optional): If True, convert to base units of length (m). Defaults to False.

    Raises:
        ValueError: If the input value is a negative number.
        pint.errors.DimensionalityError: If the input value is not a quantity of lenght.

    Returns:
        quantity (pint.Quantity): A pint quantity with units applied if it was dimensionless.
    """
    quantity = Q_(q)
    if quantity <= 0:
        raise ValueError("Quantity (length) must be positive.")
    quantity = quantity * ureg.m if quantity.dimensionless else quantity
    if quantity.check("[length]"):
        if to_base is True:
            return quantity.to_base_units()
        else:
            return quantity
    else:
        raise pint.errors.DimensionalityError(
            quantity, "a quantity of", quantity.dimensionality, ureg.mm.dimensionality
        )


def units_of_time(q: str) -> Q_:  # -> pint.Quantity:
    """
    Check that a quantity of time is compatible with NX_TIME, defaulting to s if dimensionless.
    Convert to seconds if time is passed as a fraction of it.

    Args:
        q (str): A string that can be interpreted as a pint Quantity, it can be dimensionless.

    Raises:
        ValueError: If the input value is a negative number.
        pint.errors.DimensionalityError: If the input value is not a quantity of lenght.

    Returns:
        quantity (pint.Quantity): A pint quantity in s, with units applied if it was dimensionless.
    """
    quantity = Q_(q)
    if quantity <= 0:
        raise ValueError("Quantity (time) of time must be positive.")
    quantity = quantity * ureg.s if quantity.dimensionless else quantity
    if quantity.check("[time]"):
        return quantity.to_base_units()
    else:
        raise pint.errors.DimensionalityError(
            quantity, "a quantity of", quantity.dimensionality, ureg.s.dimensionality
        )



"""
Define parameters for Electron Diffraction
"""

coordinate_frame = "mcstas"

ED_coord_system = {
    "convention": "ED",
    "origin": (0, 0, 0),
    "x": (".", "translation", "mm", [0, 1, 0]),  # (depends, type, unit, vector)
    "y": ("x", "translation", "mm", [-1, 0, 0]),
    "z": ("y", "translation", "mm", [0, 0, 1]),
}

goniometer = {
    "axes": ["alpha", "sam_z", "sam_y", "sam_x"],
    "depends": [".", "alpha", "sam_z", "sam_y"],
    "vectors": [(-1, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0)],
    "types": ["rotation", "translation", "translation", "translation"],
    "units": ["deg", "mm", "mm", "mm"],
    "offsets": [(0, 0, 0), (0, 0, 0), (0, 0, 0), (0, 0, 0)],
    "starts": None,
    "ends": None,
    "increments": None,
}

singla_1M = {
    "mode": "images",
    "description": "Dectris Singla 1M",
    "detector_type": "HPC",
    "sensor_material": "Si",
    "sensor_thickness": "0.450mm",
    "overload": 199996,
    "underload": -1,
    "pixel_size": ["0.075mm", "0.075mm"],
    "beam_center": None,
    "flatfield": None,
    "pixel_mask": None,
    "exposure_time": None,
    "image_size": [1062, 1028],  # (slow, fast)
    "axes": ["det_z"],
    "depends": ["."],
    "vectors": [0, 0, 1],
    "types": ["translation"],
    "units": ["mm"],
    "starts": None,
    "ends": None,
    "increments": 0.0,
}

module = {
    "fast_axis": [-1, 0, 0],
    "slow_axis": [0, -1, 0],
    "module_offset": "1",
}

source = {
    "name": "Diamond Light Source",
    "short_name": "DLS",
    "type": "Electron Microscope",
    "beamline_name": "eBIC",
    "probe": "electron",
}

beam = {
    "wavelength": 0.02,
    "flux": None,
}



"""
Define beamline parameters for I24 with Eiger 9M detector.
"""

goniometer_axes = {
    "axes": ["omega", "sam_z", "sam_y", "sam_x"],
    "depends": [".", "omega", "sam_z", "sam_y"],
    "vectors": [(-1, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0)],
    "types": ["rotation", "translation", "translation", "translation"],
    "units": ["deg", "mm", "mm", "mm"],
    "offsets": [(0, 0, 0), (0, 0, 0), (0, 0, 0), (0, 0, 0)],
    "starts": None,
    "ends": None,
    "increments": None,
}

eiger9M_params = {
    "mode": "images",
    "description": "Eiger 2X 9M",
    "detector_type": "Pixel",
    "sensor_material": "CdTe",
    "sensor_thickness": "0.750mm",
    "overload": 50649,  # "_dectris/countrate_correction_count_cutoff",
    "underload": -1,
    "pixel_size": ["0.075mm", "0.075mm"],
    "flatfield": "flatfield",
    "flatfield_applied": "_dectris/flatfield_correction_applied",
    "pixel_mask": "mask",
    "pixel_mask_applied": "_dectris/pixel_mask_applied",
    "image_size": [3262, 3108],  # (slow, fast)
    "axes": ["det_z"],
    "depends": ["."],
    "vectors": [0, 0, 1],
    "types": ["translation"],
    "units": ["mm"],
    "starts": None,
    "ends": None,
    "increments": None,
    "bit_depth_readout": "_dectris/bit_depth_readout",
    "detector_readout_time": "_dectris/detector_readout_time",
    "threshold_energy": "_dectris/threshold_energy",
    "software_version": "_dectris/software_version",
    "serial_number": "_dectris/detector_number",
}

eiger9M_module = {
    "fast_axis": [-1, 0, 0],
    "slow_axis": [0, -1, 0],
    "module_offset": "1",
}



"""
Command line tool to generate NXmx-like NeXus files for Electron Diffraction.
"""

import argparse
import glob
import logging
import sys
from pathlib import Path

import freephil
import h5py

from .. import get_iso_timestamp, get_nexus_filename, log
from ..beamlines.ED_params import ED_coord_system
from ..nxs_write.EDNexusWriter import ED_call_writers
from ..tools.ED_tools import extract_from_SINGLA_master, find_beam_centre
from ..tools.VDS_tools import image_vds_writer, vds_file_writer
from . import config_parser, nexus_parser, phil2dict, version_parser

logger = logging.getLogger("nexgen.EDNeXusGenerator")

ED_phil = freephil.parse(
    """
    input {
      datafiles = None
        .type = path
        .help = "List of input data files."
      coordinate_frame = mcstas
        .type = str
        .help = "Which coordinate system is being used to provide input vectors."
      vds_writer = None *dataset file
        .type = choice
        .help = "If not None, write vds along with external link to data in NeXus file, or create _vds.h5 file."
      n_imgs = None
        .type = int
        .help = "Total number of images collected."
    }

    include scope nexgen.command_line.nxs_phil.goniometer_scope

    include scope nexgen.command_line.nxs_phil.instrument_scope

    include scope nexgen.command_line.nxs_phil.detector_scope

    include scope nexgen.command_line.nxs_phil.module_scope

    include scope nexgen.command_line.nxs_phil.timestamp_scope

    include scope nexgen.command_line.nxs_phil.coord_system_scope
    """,
    process_includes=True,
)

parser = argparse.ArgumentParser(description=__doc__, parents=[version_parser])
parser.add_argument("--debug", action="store_const", const=True)


def write_from_SINGLA(args):
    cl = ED_phil.command_line_argument_interpreter()
    working_phil = ED_phil.fetch(cl.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    datafiles = [
        Path(f).expanduser().resolve()
        for f in sorted(glob.glob(params.input.datafiles))
    ]

    # Get NeXus file name
    infile = datafiles[0].parent / datafiles[0].name.replace("_data", "")
    nxsfile = get_nexus_filename(infile)

    # Reset the location of the NeXus file if -o is parsed
    if args.output:
        newdir = Path(args.output).expanduser().resolve()
        nxsfile = newdir / nxsfile.name

    # Configure logger
    logfile = nxsfile.parent / "EDnxs.log"
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for electron diffraction data from Singla.")
    logger.info(
        "Number of experiment data files in directory, linked to the Nexus file: %d"
        % len(datafiles)
    )
    logger.info("NeXus file will be saved as %s" % nxsfile)

    # Load technical info from phil parser
    coordinate_frame = params.input.coordinate_frame
    # ED_coord_system = phil2dict(params.coord_system.__dict__)
    goniometer = phil2dict(params.goniometer.__dict__)
    detector = phil2dict(params.detector.__dict__)
    module = phil2dict(params.detector_module.__dict__)
    source = phil2dict(params.source.__dict__)
    beam = phil2dict(params.beam.__dict__)
    timestamps = (
        get_iso_timestamp(params.start_time),
        get_iso_timestamp(params.end_time),
    )

    logger.info("Source information")
    logger.info(f"Facility: {source['name']} - {source['type']}.")
    logger.info(f"Beamline / instrument: {source['beamline_name']}")

    logger.warning(
        f"Coordinate frame of input arrays currently set to {coordinate_frame}."
        "If that is not the case, please indicate the correct one with --coord-frame."
        "For more information, see the help message."
    )

    logger.warning(
        "Have you checked the coordinate system convention?\n"
        "If no new values have been passed for coord_system.origin or coord_system.vectors the following convention will be applied:\n"
        f"{ED_coord_system}"
    )

    # If anything has been passed regarding the new coordinate system convention
    # overwrite the existing dictionary
    if params.coord_system.convention:
        logger.info(
            f"New coordinate system convention: {params.coord_system.convention}."
        )
        ED_coord_system["convention"] = params.coord_system.convention

    if params.coord_system.origin:
        logger.info(
            f"New value for coordinate system found: {params.coord_system.origin}."
        )
        ED_coord_system["origin"] = tuple(params.coord_system.origin)

    if params.coord_system.vectors:
        from .. import split_arrays

        # Note: setting to coordinate frame to avoid any conversions. FIXME
        vectors = split_arrays(["x", "y", "z"], params.coord_system.vectors)
        logger.info(
            f"New vectors defined for {params.coord_system.convention} coordinate system."
        )
        ED_coord_system["x"] = (".", "translation", "mm", vectors["x"])
        ED_coord_system["y"] = ("x", "translation", "mm", vectors["y"])
        ED_coord_system["z"] = ("y", "translation", "mm", vectors["z"])

    if args.master:
        master = Path(args.master).expanduser().resolve()
        logger.info(
            f"Looking through Dectris master file to extract at least mask and flatfield."
        )
        detector.update(extract_from_SINGLA_master(master))

        # Calculate beam centre if missing
        if detector["beam_center"] is None:
            detector["beam_center"] = find_beam_centre(master, datafiles[0])
            if detector["beam_center"] is None:
                detector["beam_center"] = (0, 0)
                logger.warning(
                    f"Unable to calculate beam centre. It has been set to {detector['beam_center']}."
                )
            else:
                logger.info(f"Calculated beam centre to be {detector['beam_center']}.")

    # Start writing
    logger.info("Start writing NeXus file ...")
    try:
        with h5py.File(nxsfile, "x") as nxs:
            ED_call_writers(
                nxs,
                datafiles,
                goniometer,
                detector,
                module,
                source,
                beam,
                ED_coord_system,
                coordinate_frame=coordinate_frame,
                n_images=params.input.n_imgs,
                timestamps=timestamps,
            )

            if params.input.vds_writer == "dataset":
                nimages = (
                    nxs["/entry/instrument/detector/detectorSpecific/nimages"][()]
                    if params.input.n_imgs is None
                    else params.input.n_imgs
                )
                logger.info("Calling VDS writer ...")
                image_vds_writer(
                    nxs,
                    (int(nimages), *detector["image_size"]),
                    entry_key="/entry/data/data",
                )
            elif params.input.vds_writer == "file":
                nimages = (
                    nxs["/entry/instrument/detector/detectorSpecific/nimages"][()]
                    if params.input.n_imgs is None
                    else params.input.n_imgs
                )
                logger.info(
                    "Calling VDS writer to write a Virtual Dataset file and relative link."
                )
                vds_file_writer(nxs, datafiles, (int(nimages), *detector["image_size"]))
            else:
                logger.info("VDS won't be written.")

        logger.info("NeXus file written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(f"An error occurred and {nxsfile} couldn't be written correctly.")


subparsers = parser.add_subparsers(
    help="Define what to do with input file depending on their origin.",
    required=True,
    dest="sub-command",
)

singla_parser = subparsers.add_parser(
    "singla",
    description=("Trigger NeXus file writing for Singla data."),
    parents=[nexus_parser, config_parser],
)
singla_parser.add_argument(
    "-m",
    "--master",
    type=str,
    help="HDF5 master file written by Singla detector.",
)
singla_parser.add_argument(
    "-o",
    "--output",
    type=str,
    help="Output directory if different from location of data files.",
)
singla_parser.set_defaults(func=write_from_SINGLA)


def main():
    args = parser.parse_args()
    args.func(args)



"""
Experiment definitions for SSX collections:
    - extruder
    - fixed target
    - 3D grid scan
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np

from ..nxs_write.NexusWriter import ScanReader
from . import PumpProbe
from .SSX_chip import Chip, compute_goniometer, read_chip_map

__all__ = ["run_extruder", "run_fixed_target", "run_3D_grid_scan"]

# Define a logger object
logger = logging.getLogger("nexgen.SSX.run_expt")


def run_extruder(
    goniometer: Dict[str, List],
    num_imgs: int,
    pump_probe: PumpProbe,
) -> Tuple[Dict]:
    """Run the goniometer computations for an extruder experiment.

    Args:
        goniometer (Dict[str, List]): Goniometer definition.
        num_imgs (int): Total number of images.
        pump_probe (PumpProbe): Pump probe parameters.

    Returns:
        Tuple[Dict]:
            goniometer: updated goniometer dictionary with actual values from the scan.
            OSC: dictionary with oscillation scan axis values.
            pump_info: updated pump probe information.
    """
    logger.info("Running an extruder experiment.")

    logger.debug("All axes are fixed, setting increments to 0.0 and starts == ends.")
    goniometer["increments"] = len(goniometer["axes"]) * [0.0]
    if goniometer["starts"] is None:
        goniometer["starts"] = len(goniometer["axes"]) * [0.0]
    goniometer["ends"] = goniometer["starts"]

    OSC, TRANSL = ScanReader(goniometer, n_images=int(num_imgs))
    del TRANSL

    pump_info = pump_probe.to_dict()

    return goniometer, OSC, pump_info


def run_fixed_target(
    goniometer: Dict[str, List],
    chip_info: Dict[str, List],
    chipmap: Path | str,
    pump_probe: PumpProbe,
    osc_axis: str = "omega",
) -> Tuple[Dict]:
    """Run the goniometer computations for a fixed-target experiment.

    Args:
        goniometer (Dict[str, List]): Goniometer definition.
        chip_info (Dict[str, List]): General information about the chip: number and size of blocks, size and step of each window, start positions, number of exposures.
        chipmap (Path | str): Path to .map file. If None is passed, assumes a fullchip.
        pump_probe (PumpProbe): Pump probe parameters.
        osc_axis (str, optional): Oscillation axis. Defaults to omega.

    Returns:
        Tuple[Dict]:
            goniometer: updated goniometer dictionary with actual values from the scan.
            OSC: dictionary with oscillation scan axis values.
            TRANSL: dictionary with grid scan values.
            pump_info: updated pump probe information.
    """
    logger.info("Running a fixed target experiment.")

    # Check that the chip dict has been passed, raise error is not
    if chip_info is None:
        logger.error("No chip_dict found.")
        raise ValueError(
            "No information about the FT chip has been passed. \
            Impossible to determine scan parameters. NeXus file won't be written."
        )

    # Define chip
    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"][1], chip_info["Y_NUM_STEPS"][1]],
        step_size=[chip_info["X_STEP_SIZE"][1], chip_info["Y_STEP_SIZE"][1]],
        num_blocks=[chip_info["X_NUM_BLOCKS"][1], chip_info["Y_NUM_BLOCKS"][1]],
        block_size=[chip_info["X_BLOCK_SIZE"][1], chip_info["Y_BLOCK_SIZE"][1]],
        start_pos=[
            chip_info["X_START"][1],
            chip_info["Y_START"][1],
            chip_info["Z_START"][1],
        ],
    )

    # Read chip map
    blocks = read_chip_map(
        chipmap,
        chip.num_blocks[0],
        chip.num_blocks[1],
    )

    # Set step size as increment for grid scan axes and set everything else to 0
    l = len(goniometer["axes"])
    # Temporary workaround for I19 scripts which saves an increment for phi in meta file.
    goniometer["increments"] = l * [0.0]
    Yidx, Xidx = (
        goniometer["axes"].index("sam_y"),
        goniometer["axes"].index("sam_x"),
    )
    goniometer["increments"][Xidx] = chip.step_size[0]
    goniometer["increments"][Yidx] = chip.step_size[1]

    # Sanity check
    if goniometer["starts"] is not None and goniometer["starts"] != goniometer["ends"]:
        logger.debug(
            "Sanity check in case the meta file has a wrong omega/phi end value."
        )
        logger.debug(
            "In this application there is no rotation so setting end values to the same as start."
        )
        goniometer["ends"] = [s for s in goniometer["starts"]]

    # Calculate scan start/end positions on chip
    if list(blocks.values())[0] == "fullchip":
        logger.info("Full chip: all the blocks will be scanned.")
        from .SSX_chip import fullchip_blocks_conversion

        start_pos, end_pos = compute_goniometer(chip, goniometer["axes"], full=True)
        start_pos = fullchip_blocks_conversion(start_pos, chip)
        end_pos = fullchip_blocks_conversion(end_pos, chip)
    else:
        logger.info(f"Scanning blocks: {list(blocks.keys())}.")
        start_pos, end_pos = compute_goniometer(chip, goniometer["axes"], blocks=blocks)

    # Iterate over blocks to calculate scan points
    OSC = {osc_axis: np.array([])}
    TRANSL = {"sam_y": np.array([]), "sam_x": np.array([])}
    for _s, _e in zip(start_pos.items(), end_pos.items()):
        # Determine wheter it's an up or down block
        col = (
            int(_e[0]) // chip.num_blocks[0]
            if int(_e[0]) % chip.num_blocks[0] != 0
            else (int(_e[0]) // chip.num_blocks[0]) - 1
        )
        # Get the values
        if goniometer["starts"] is None:
            s = _s[1]
        else:
            s = [
                goniometer["starts"][i] if i not in [Xidx, Yidx] else _s[1][i]
                for i in range(l)
            ]
        if goniometer["ends"] is None:
            e = _e[1]
        else:
            e = [
                goniometer["ends"][i] if i not in [Xidx, Yidx] else _e[1][i]
                for i in range(l)
            ]
        goniometer["starts"] = s
        # Workaround for scanspec issue (we don't want to write the actual end of the chip)
        if col % 2 == 0:
            goniometer["ends"] = [
                end - inc for end, inc in zip(e, goniometer["increments"])
            ]
        else:
            goniometer["ends"] = [
                e[i] if i != Xidx else e[i] - goniometer["increments"][i]
                for i in range(len(e))
            ]
        logger.debug(
            "Current block: \n"
            f"Starts: {goniometer['starts']} \n"
            f"Ends: {goniometer['ends']} \n"
            f"Incs: {goniometer['increments']}"
        )
        osc, transl = ScanReader(
            goniometer,
            n_images=(
                chip.num_steps[1],
                chip.num_steps[0],
            ),
            osc_axis=osc_axis,
        )
        OSC[osc_axis] = np.append(OSC[osc_axis], osc[osc_axis])
        TRANSL["sam_y"] = np.append(TRANSL["sam_y"], np.round(transl["sam_y"], 3))
        TRANSL["sam_x"] = np.append(TRANSL["sam_x"], np.round(transl["sam_x"], 3))

    N = int(chip_info["N_EXPOSURES"][1])
    if N > 1:
        # Repeat each position N times
        OSC = {k: [val for val in v for _ in range(N)] for k, v in OSC.items()}
        TRANSL = {k: [val for val in v for _ in range(N)] for k, v in TRANSL.items()}

    logger.info(f"Each position has been collected {N} times.")
    logger.info(f"Pump repeat setting: {chip_info['PUMP_REPEAT'][1]}.")
    pump_info = pump_probe.to_dict()
    pump_info["repeat"] = int(chip_info["PUMP_REPEAT"][1])
    pump_info["n_exposures"] = N

    return goniometer, OSC, TRANSL, pump_info


def run_3D_grid_scan(
    goniometer: Dict[str, List],
    chip_info: Dict[str, List],
    chipmap: Path | str,
    pump_probe: PumpProbe,
    osc_axis: str = "omega",
) -> Tuple[Dict]:
    """_summary_

    Args:
        goniometer (Dict[str, List]): _description_
        chip_info (Dict[str, List]): _description_
        chipmap (Path | str): _description_
        pump_probe (PumpProbe): _description_
        osc_axis (str, optional): _description_

    Returns:
        Tuple[Dict]:
            goniometer: updated goniometer dictionary with actual values from the scan.
            OSC: dictionary with oscillation scan axis values
            TRANSL: dictionary with grid scan values
            pump_info: updated pump probe information
    """
    logger.info("Running a 3D grid scan experiment.")

    N = int(chip_info["N_EXPOSURES"][1])

    pump_info = pump_probe.to_dict()
    pump_info["repeat"] = int(chip_info["PUMP_REPEAT"][1])
    pump_info["n_exposures"] = N
    return None, None, None, pump_info



"""
Create a NeXus file for time-resolved collections on I19-2.
"""

import glob
import logging
from collections import namedtuple
from pathlib import Path
from typing import List, Tuple

import h5py

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write import calculate_scan_range, find_number_of_images, find_osc_axis
from ..nxs_write.NexusWriter import call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry
from ..tools.MetaReader import overwrite_beam, update_detector_axes, update_goniometer
from ..tools.VDS_tools import image_vds_writer
from . import eiger_meta_links as dset_links
from . import source
from .I19_2_params import (
    eiger4M_module,
    eiger4M_params,
    goniometer_axes,
    tristan10M_module,
    tristan10M_params,
)

# Define a logger object
logger = logging.getLogger("nexgen.I19-2_NeXus")

# Tristan mask and flatfield files
maskfile = "Tristan10M_mask_with_spec.h5"
flatfieldfile = "Tristan10M_flat_field_coeff_with_Mo_17.479keV.h5"

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize dictionaries
goniometer = {}
detector = {}
module = {}
beam = {"flux": None}
attenuator = {}

tr_collect = namedtuple(
    "tr_collect",
    [
        "meta_file",
        "detector_name",
        "exposure_time",
        "transmission",
        "wavelength",
        "beam_center",
        "start_time",
        "stop_time",
        "scan_axis",
    ],
)

tr_collect.__doc__ = """Parameters passed as input from the beamline."""
tr_collect.meta_file.__doc__ = "Path to _meta.h5 file."
tr_collect.detector_name.__doc__ = "Name of the detector in use for current experiment."
tr_collect.exposure_time.__doc__ = "Exposure time, in s."
tr_collect.transmission.__doc__ = "Attenuator transmission, in %."
tr_collect.wavelength.__doc__ = "Incident beam wavelength, in A."
tr_collect.beam_center.__doc__ = "Beam center (x,y) position, in pixels."
tr_collect.start_time.__doc__ = "Collection start time."
tr_collect.stop_time.__doc__ = "Collection end time."
tr_collect.scan_axis.__doc__ = "Rotation scan axis. Must be passed for Tristan."


def tristan_writer(
    master_file: Path,
    TR: namedtuple,
    timestamps: Tuple[str, str] = (None, None),
    axes_pos: List[namedtuple] = None,
    det_pos: List[namedtuple] = None,
):
    """
    A function to call the nexus writer for Tristan 10M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
        axes_pos (List[namedtuple], optional): List of (axis_name, start, end) values for the goniometer, passed from command line. Defaults to None.
        det_pos (List[namedtuple], optional): List of (axis_name, start) values for the detector, passed from command line. Defaults to None.
    """
    # Add tristan mask and flatfield files
    detector["pixel_mask"] = "Tristan10M_mask_with_spec.h5"
    detector["flatfield"] = "Tristan10M_flat_field_coeff_with_Mo_17.479keV.h5"

    # Update axes
    # Goniometer
    l = len(goniometer["axes"])
    goniometer["starts"] = l * [0.0]
    goniometer["increments"] = l * [0.0]
    goniometer["ends"] = l * [0.0]
    for ax in axes_pos:
        idx = goniometer["axes"].index(ax.id)
        goniometer["starts"][idx] = ax.start
        goniometer["ends"][idx] = ax.end

    # Detector
    detector["starts"] = detector["ends"] = len(detector["axes"]) * [0.0]
    for dax in det_pos:
        idx = detector["axes"].index(dax.id)
        detector["starts"][idx] = dax.start

    # Identify scan axis and calculate scan range
    if TR.scan_axis is None:
        scan_axis = find_osc_axis(
            goniometer["axes"],
            goniometer["starts"],
            goniometer["ends"],
            goniometer["types"],
        )
    else:
        scan_axis = TR.scan_axis
    scan_idx = goniometer["axes"].index(scan_axis)
    OSC = {scan_axis: (goniometer["starts"][scan_idx], goniometer["ends"][scan_idx])}

    logger.info("--- COLLECTION SUMMARY ---")
    logger.info("Source information")
    logger.info(f"Facility: {source['name']} - {source['type']}.")
    logger.info(f"Beamline: {source['beamline_name']}")

    logger.info(f"Incident beam wavelength: {beam['wavelength']}")
    logger.info(f"Attenuation: {attenuator['transmission']}")

    logger.info("Goniometer information")
    logger.info(f"Scan axis is: {scan_axis}")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {detector['beam_center']}.")

    logger.info(f"Timestamps recorded: {timestamps}")

    # Write
    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))

            call_writers(
                nxsfile,
                [TR.meta_file],
                coordinate_frame,
                (detector["mode"], None),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                sample_depends_on=scan_axis,
            )

            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise


def eiger_writer(
    master_file: Path,
    TR: namedtuple,
    timestamps: Tuple[str, str] = (None, None),
):
    """
    A function to call the nexus writer for Eiger 2X 4M detector.
    It requires the informatin contained inside the meta file to work correctly.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).

    Raises:
        IOError: If the axes positions can't be read from the metafile (missing config or broken links).
    """
    # Find datafiles
    logger.info("Looking for data files ...")
    filename_template = (
        TR.meta_file.parent / TR.meta_file.name.replace("meta", f"{6*'[0-9]'}")
    ).as_posix()
    filenames = [
        Path(f).expanduser().resolve() for f in sorted(glob.glob(filename_template))
    ]
    # Calculate total number of images
    n_frames = find_number_of_images(filenames)
    logger.info(
        f"Found {len(filenames)} files in directory, containing {n_frames} images."
    )

    # Update axes
    with h5py.File(TR.meta_file, "r") as mh:
        update_goniometer(mh, goniometer)
        update_detector_axes(mh, detector)
        logger.info(
            "Goniometer and detector axes positions have been updated with values from the meta file."
        )
        if beam["wavelength"] is None:
            logger.info(
                "Wavelength has't been passed by user. Looking for it in the meta file."
            )
            overwrite_beam(mh, TR.detector_name, beam)
        if detector["beam_center"] is None:
            logger.info(
                "Beam center position has't been passed by user. Looking for it in the meta file."
            )
            from ..tools.Metafile import DectrisMetafile

            meta = DectrisMetafile(mh)
            detector["beam_center"] = meta.get_beam_center()

    # Check that axes have been updated
    if goniometer["starts"] is None:
        raise IOError("Goniometer axes values couldn't be read from meta file.")
        # FOr now. If it doesn't work, more than likely meta is broken but axes can be passed.
    if detector["starts"] is None:
        raise IOError("Detector axes values couldn't be read from meta file.")

    scan_axis = find_osc_axis(
        goniometer["axes"],
        goniometer["starts"],
        goniometer["ends"],
        goniometer["types"],
    )
    scan_idx = goniometer["axes"].index(scan_axis)
    OSC = calculate_scan_range(
        [scan_axis],
        [goniometer["starts"][scan_idx]],
        [goniometer["ends"][scan_idx]],
        axes_increments=[goniometer["increments"][scan_idx]],
        # n_images=n_frames,
        rotation=True,
    )
    TRANSL = None

    logger.info("--- COLLECTION SUMMARY ---")
    logger.info("Source information")
    logger.info(f"Facility: {source['name']} - {source['type']}.")
    logger.info(f"Beamline: {source['beamline_name']}")

    logger.info(f"Incident beam wavelength: {beam['wavelength']}")
    logger.info(f"Attenuation: {attenuator['transmission']}")

    logger.info("Goniometer information")
    logger.info(f"Scan axis is: {scan_axis}")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {detector['beam_center']}.")

    logger.info(f"Timestamps recorded: {timestamps}")

    # Write
    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))

            call_writers(
                nxsfile,
                filenames,
                coordinate_frame,
                (detector["mode"], n_frames),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=TRANSL,
                metafile=TR.meta_file,
                link_list=dset_links,
                sample_depends_on=scan_axis,
            )

            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))

            # Write VDS
            image_vds_writer(nxsfile, (int(n_frames), *detector["image_size"]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise


def nexus_writer(**params):
    """
    Gather all parameters from the beamline and call the NeXus writers.
    """
    TR = tr_collect(
        meta_file=Path(params["meta_file"]).expanduser().resolve(),
        detector_name=params["detector_name"],
        exposure_time=params["exposure_time"],
        transmission=params["transmission"],
        wavelength=params["wavelength"],
        beam_center=params["beam_center"],
        start_time=params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")  #
        if params["start_time"]
        else None,
        stop_time=params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")  #
        if params["stop_time"]
        else None,
        scan_axis=params["scan_axis"] if params["scan_axis"] else None,
    )

    # Check that the new NeXus file is to be written in the same directory
    if params["outdir"]:
        wdir = Path(params["outdir"]).expanduser().resolve()
    else:
        wdir = TR.meta_file.parent

    # Define a file handler
    logfile = wdir / "I19_2_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for beamline I19-2 at DLS.")
    logger.info(f"Detector in use for this experiment: {TR.detector_name}.")
    logger.info(f"Current collection directory: {TR.meta_file.parent}")

    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % TR.meta_file.name)
    # Get NeXus filename
    master_file = get_nexus_filename(TR.meta_file)
    master_file = wdir / master_file.name
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get timestamps in the correct format if they aren't already
    timestamps = (
        get_iso_timestamp(TR.start_time),
        get_iso_timestamp(TR.stop_time),
    )

    # logger.info("Load goniometer from I19-2.")
    for k, v in goniometer_axes.items():
        goniometer[k] = v

    # Fill in a few dictionaries
    attenuator["transmission"] = TR.transmission if TR.transmission else None

    beam["wavelength"] = TR.wavelength

    if "tristan" in TR.detector_name.lower():
        for k, v in tristan10M_params.items():
            detector[k] = v
        if params["gonio_pos"] is None or params["det_pos"] is None:
            logger.error("Please pass the axes positions for a Tristan collection.")
        if TR.scan_axis is None:
            logger.warning("No scan axis has been specified.")
        for k, v in tristan10M_module.items():
            module[k] = v
    else:
        for k, v in eiger4M_params.items():
            detector[k] = v
        for k, v in eiger4M_module.items():
            module[k] = v

    detector["exposure_time"] = TR.exposure_time
    detector["beam_center"] = TR.beam_center

    if "eiger" in TR.detector_name.lower() and params["serial"] is False:
        eiger_writer(master_file, TR, timestamps)
    elif "tristan" in TR.detector_name.lower():
        tristan_writer(
            master_file, TR, timestamps, params["gonio_pos"], params["det_pos"]
        )



"""
Tools to extract metadata for Electron Diffraction.
"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Tuple

import h5py
import hdf5plugin  # noqa: F401
import numpy as np
from numpy.typing import ArrayLike

try:
    # Only for Python version >= 3.8
    from functools import cached_property
except ImportError:
    # Compatibility for earlier Python versions
    import functools

    def cached_property(func):
        @property
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper_decorator


class SinglaMaster:
    """
    Describes a master file for a Dectris Singla detector.
    """

    @staticmethod
    def isDectrisSingla(filename):
        with h5py.File(filename, "r") as fh:
            description = fh["/entry/instrument/detector/description"][()]
        if b"SINGLA" in description.upper():
            return True
        return False

    def __init__(self, handle: h5py.File):
        self._handle = handle

    def __len__(self):
        return len(self._handle)

    def __str__(self):
        return f"File {self._handle.filename} opened in '{self._handle.mode}' mode."

    def __getitem__(self, key: str) -> h5py.Group | h5py.Dataset:
        return self._handle[key]

    @cached_property
    def walk(self) -> List[str]:
        obj_list = []
        self._handle.visit(obj_list.append)
        return obj_list

    def get_number_of_images(self) -> int:
        _loc = [obj for obj in self.walk if "nimages" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_number_of_triggers(self) -> int:
        _loc = [obj for obj in self.walk if "ntriggers" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_trigger_mode(self) -> str:
        _loc = [obj for obj in self.walk if "trigger_mode" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_mask(self) -> Tuple[bool, ArrayLike]:
        M = [obj for obj in self.walk if "pixel_mask" in obj]
        if len(M) > 0:
            mask_path = [_loc for _loc in M if _loc.split("/")[-1] == "pixel_mask"]
            mask_applied_path = [_loc for _loc in M if "applied" in _loc]
            if len(mask_applied_path) == 0:
                return (None, self.__getitem__(mask_path[0])[()])
            if len(mask_path) == 0:
                return (self.__getitem__(mask_applied_path[0])[()], None)
            return (
                self.__getitem__(mask_applied_path[0])[()],
                self.__getitem__(mask_path[0])[()],
            )
        return (None, None)

    def get_flafield(self) -> Tuple[bool, ArrayLike]:
        F = [obj for obj in self.walk if "flatfield" in obj]
        if len(F) > 0:
            flatfield_path = [_loc for _loc in F if _loc.split("/")[-1] == "flatfield"]
            flatfield_applied_path = [_loc for _loc in F if "applied" in _loc]
            if len(flatfield_applied_path) == 0:
                return (None, self.__getitem__(flatfield_path[0])[()])
            if len(flatfield_path) == 0:
                return (self.__getitem__(flatfield_applied_path[0])[()], None)
            return (
                self.__getitem__(flatfield_applied_path[0])[()],
                self.__getitem__(flatfield_path[0])[()],
            )
        return (None, None)

    def get_bit_bepth_readout(self) -> int:
        _loc = [obj for obj in self.walk if "bit_depth_readout" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]  # type SCALAR

    def get_bit_bepth_image(self) -> int:
        _loc = [obj for obj in self.walk if "bit_depth_image" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_detector_number(self) -> str:
        _loc = [obj for obj in self.walk if "detector_number" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_detector_readout_time(self) -> float:
        _loc = [obj for obj in self.walk if "detector_readout_time" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_photon_energy(self) -> float:
        _loc = [obj for obj in self.walk if "photon_energy" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_countrate_correction(self) -> int:
        _loc = [obj for obj in self.walk if "countrate_correction_applied" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]

    def get_software_version(self) -> bytes:
        _loc = [obj for obj in self.walk if "software_version" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[()]


def extract_from_SINGLA_master(master: Path | str) -> Dict[str, Any]:
    """
    Extracts mask, flatfield and any other information relative to the detector \
    from a Singla master file.

    Args:
        master (Path | str): Path to Singla master file.

    Returns:
        Dict[str, Any]: Dictionary of information relative to the detector.
    """
    import logging

    logger = logging.getLogger("nexgen.EDtools.Singla")
    logger.setLevel(logging.DEBUG)

    if SinglaMaster.isDectrisSingla(master) is False:
        logger.warning(f"The file {master} is the wrong format.")
        return

    D = {}
    with h5py.File(master, "r") as fh:
        singla = SinglaMaster(fh)
        MASK = singla.get_mask()
        if MASK[1] is not None:
            logger.info("Pixel_mask has been found in master file.")
        FF = singla.get_flafield()
        if FF[1] is not None:
            logger.info("Flatfield has been found in master file.")
        D["pixel_mask"] = MASK[1]
        D["pixel_mask_applied"] = MASK[0]
        D["flatfield"] = FF[1]
        D["flatfield_applied"] = FF[0]
        D["software_version"] = singla.get_software_version()
        D["bit_depth_readout"] = singla.get_bit_bepth_readout()
        D["detector_number"] = singla.get_detector_number()
        D["detector_readout_time"] = singla.get_detector_readout_time()
        D["photon_energy"] = singla.get_photon_energy()

    return D


def centroid_max(image: ArrayLike) -> Tuple[float, float]:
    """
    Find the centre of gravity of the maximum pixels.

    Args:
        image (ArrayLike): Pixel image.

    Returns:
        Tuple[float, float]: Centroid (x,y) position.
    """

    y, x = np.where(image == np.amax(image))
    return np.mean(x), np.mean(y)


def find_beam_centre(
    master: Path | str, data: Path | str, data_entry_key: str = "/entry/data/data"
) -> Tuple[float, float]:
    """
    Calculate the beam center position for Electron Diffraction data collected on Singla detector.

    Args:
        master (Path | str): Path to Singla master file.
        data (Path | str): Path to data file.
        data_entry_key (str, optional): Key for the location of the images inside the Singla data file. Defaults to "/entry/data/data".

    Returns:
        fast, slow (Tuple[float, float]): Beam center position (fast, slow) on the detector. \
            None if the pixel_mask can't be found.
    """

    with h5py.File(master, "r") as fh:
        singla = SinglaMaster(fh)
        pixel_mask = singla.get_mask()[1]

    # If the module gap is unmasked we would get bad results
    if pixel_mask is None:
        return None

    # Set the ROI to be +/- 100 pixels around the image centre
    yc, xc = (e // 2 for e in pixel_mask.shape)
    x0 = xc - 100
    x1 = xc + 100
    y0 = yc - 100
    y1 = yc + 100

    # Bool selection for masked pixels in the ROI
    pixel_mask = pixel_mask[y0:y1, x0:x1] == 1

    images = []
    with h5py.File(data, "r") as fh:
        data = fh[data_entry_key]
        num_images = data.shape[0]
        for i in range(0, num_images, num_images // min(num_images, 10)):
            image = data[i, y0:y1, x0:x1]
            image[pixel_mask] = 0
            images.append(image)

    beam_centres = [centroid_max(im) for im in images]
    x, y = zip(*beam_centres)

    # For robustness against blank images, remove any value more than 5 px
    # from the median
    med_x = np.median(x)
    med_y = np.median(y)
    x = [e for e in x if abs(e - med_x) < 5]
    y = [e for e in y if abs(e - med_y) < 5]

    # Correct for offset of the ROI and shift to centre pixel
    fast = xc - 100 + np.mean(x) + 0.5
    slow = yc - 100 + np.mean(y) + 0.5

    return fast, slow



"""
Writer for NXmx format NeXus files for Electron Diffraction.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import h5py

from .. import reframe_arrays
from . import find_number_of_images
from .NexusWriter import ScanReader
from .NXclassWriters import (
    write_NXcoordinate_system_set,
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXnote,
    write_NXsample,
    write_NXsource,
)

logger = logging.getLogger("nexgen.ED_Writer")
logger.setLevel(logging.DEBUG)


# Write NeXus base classes for ED.
def ED_call_writers(
    nxsfile: h5py.File,
    datafiles: List[Path | str],
    goniometer: Dict[str, Any],
    detector: Dict[str, Any],
    module: Dict[str, Any],
    source: Dict[str, Any],
    beam: Dict[str, Any],
    ED_coord_system: Dict[str, Any],
    coordinate_frame: str = "mcstas",
    n_images: int = None,
    timestamps: List | Tuple = None,
    notes: Dict[str, Any] = None,
):
    """
    Write a new NXmx format-like NeXus file for Electron Diffraction data.
    This function performs a few checks on the coordinate frame of the input vectors \
    and then calls the writers for the relevant NeXus base classes.

    Args:
        nxsfile (h5py.File): Handle to NeXus file to be written.
        datafiles (List[Path  |  str]): List of at least 1 Path object to a HDF5 data file.
        goniometer (Dict[str, Any]): Goniometer geometry description.
        detector (Dict[str, Any]): Detector specific parameters and its axes.
        module (Dict[str, Any]): Geometry and description of detector module.
        source (Dict[str, Any]): Facility information.
        beam (Dict[str, Any]): Beam properties.
        ED_coord_system (Dict[str, Any]): Definition of the current coordinate frame for ED. \
            It should at least contain the convention, origin and base vectors.
        coordinate_frame (str, optional): Coordinate system being used. Defaults to "mcstas".
        n_images (int, optional): _description_. Defaults to None.
        timestamps (List | Tuple, optional): Start and end time of the collection, if known. \
            Preferably passed as datetime.datetime. Defaults to None.
        notes (Dict[str, Any], optional): Any useful information/comment about the collection. \
            The keys of the dictionaries will be the dataset names and the values the data. Defaults to None.
    """
    logger.info("Call the NXclass writers for Electron Diffraction data.")

    # For the moment since there's no attenuator just set to none
    attenuator = {"transmission": None}

    # Deal with vecotrs/offsets/whatever
    reframe_arrays(
        goniometer,
        detector,
        module,
        coordinate_frame,
        ED_coord_system,
    )

    # Check that filenames are paths
    if all(isinstance(f, Path) for f in datafiles) is False:
        datafiles = [Path(f).expanduser().resolve() for f in datafiles]

    # Define entry_key if dealing with singla detector
    data_entry_key = (
        "/entry/data/data" if "SINGLA" in detector["description"].upper() else "data"
    )

    # If n_images is not passed, calculate it from data files
    if not n_images:
        n_images = find_number_of_images(datafiles, data_entry_key)
        logger.info(f"Total number of images: {n_images}.")

    if goniometer["ends"] is None:
        logger.info(
            "Goniometer end position has not been passed."
            "The value for the rotation axis will be calculated from the number of images."
        )
        ax_idx = [
            goniometer["increments"].index(i)
            for i in goniometer["increments"]
            if i != 0
        ][0]
        end = goniometer["starts"][ax_idx] + goniometer["increments"][ax_idx] * n_images
        goniometer["ends"] = [
            end if i == ax_idx else 0.0 for i in range(len(goniometer["axes"]))
        ]
        logger.info(f"Goniometer end positions set to {goniometer['ends']}")

    # Define data_type
    data_type = ("images", n_images)

    # Calculate scan
    OSC, _ = ScanReader(
        goniometer, n_images=n_images
    )  # No grid scan, can be added if needed at later time
    logger.info(f"Rotation scan axis: {list(OSC.keys())[0]}.")
    logger.info(
        f"Scan from {list(OSC.values())[0][0]} to {list(OSC.values())[0][-1]}.\n"
    )

    # NXentry: /entry
    write_NXentry(nxsfile)

    # NXcoordinate_system_set: /entry/coordinate_system_set
    base_vectors = {k: ED_coord_system.get(k) for k in ["x", "y", "z"]}
    write_NXcoordinate_system_set(
        nxsfile, ED_coord_system["convention"], base_vectors, ED_coord_system["origin"]
    )

    # NXdata: /entry/data
    write_NXdata(
        nxsfile, datafiles, goniometer, data_type, OSC, entry_key=data_entry_key
    )

    # NXinstrument: /entry/instrument
    write_NXinstrument(
        nxsfile,
        beam,
        attenuator,
        source,
    )

    # NXdetector: /entry/instrument/detector
    write_NXdetector(
        nxsfile,
        detector,
        data_type,
    )

    # NXmodule: /entry/instrument/detector/module
    write_NXdetector_module(
        nxsfile,
        module,
        detector["image_size"],
        detector["pixel_size"],
        beam_center=detector["beam_center"],
    )

    # NXsource: /entry/source
    write_NXsource(nxsfile, source)

    # NXsample: /entry/sample
    write_NXsample(
        nxsfile,
        goniometer,
        data_type,
        OSC,
    )

    # NXdatetime: /entry/start_time and /entry/stop_time
    if timestamps:
        write_NXdatetime(nxsfile, timestamps)

    # NXnote: /entry/notes
    if notes:
        write_NXnote(nxsfile, "/entry/notes", notes)
